
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from scipy.stats import pearsonr
import pickle
from types import SimpleNamespace
from sklearn.linear_model import LinearRegression
import importlib # reimport package after edits: importlib.reload(bldr)

# from courtney-barrer's fork of pyzelda
import pyzelda.zelda as zelda
import pyzelda.ztools as ztools
import pyzelda.utils.aperture as aperture
import pyzelda.utils.imutils as imutils
from baldrapp.common import phasescreens as ps
from baldrapp.common import utilities as util
from baldrapp.common import baldr_core as bldr

from baldrapp.common.baldr_core import StrehlModel 

"""
Most of the calibration of the Strehl model done in this script has been generalized to a function in the baldr_core module:

bldr.calibrate_strehl_model( zwfs_ns, save_results_path = 'path/to/save/results', train_fraction = 0.6, correlation_threshold = 0.5, \
    number_of_screen_initiations = 50, scrn_scaling_grid = np.logspace(-1,0.2,5) )    
    

The end of this script does some analysis on the residual of the predicted optical gain from the Strehl model (b=b0 * sqrt(S))
in different strehl regions and as a function of the radial profile of the optical gain.

results were saved in sydney_test overleaf project.
"""



def compute_correlation_map(intensity_frames, strehl_ratios):
    # intensity_frames: k x N x M array (k frames of N x M pixels)
    # strehl_ratios: k array (Strehl ratio for each frame)
    
    k, N, M = intensity_frames.shape
    correlation_map = np.zeros((N, M))
    
    for i in range(N):
        for j in range(M):
            pixel_intensity_series = intensity_frames[:, i, j]
            correlation_map[i, j], _ = pearsonr(pixel_intensity_series, strehl_ratios)
    
    return correlation_map





# Function to load the model from a pickle file
def load_model_from_pickle(filename):
    """
    Loads the StrehlModel object from a pickle file.
    
    Args:
        filename (str): The file path from where the model should be loaded.
    
    Returns:
        StrehlModel: The loaded StrehlModel instance.
    """
    with open(filename, 'rb') as file:
        model = pickle.load(file)
        
    return model

# Function to save the model using pickle
# def save_model(filename, model):
#     with open(filename, 'wb') as f:
#         pickle.dump(model, f)

# # Function to load the model using pickle
# def load_model(filename):
#     with open(filename, 'rb') as f:
#         return pickle.load(f)
    
# create phase screen 

# function to remove N modes from the phase screen

# create vibration spectrum 

# create time series of the vibration spectrum

# for a non-aberrated system 
#   calculate the non-observable parameters 
#       - optical gain (b)
#   simulate the observable parameters
#       - ZWFS signal with and without phasemask
#   find a way to estimate b0. Problem is that we don't know the internal aberrations.
#   The general model is:
#       I0 - N0 = abs(psi_A)**2 + abs(psi_R)**2 + 2 * abs(psi_A) * abs(psi_R) * np.cos(  mu ) - abs(psi_A)**2    
#   Problem statement:
#   We have measurements of two images (intensities in a NxM camera defined by x,y pixel coordinates)  I0(x,y), N0(x,y). 
#   I0(x,y) is the intensity with a phase shifting phase mask inserted at focus and N0(x,y) is the intensity without 
#   the phasemask which is essentailly a uniformly illuminated pupil. The difference  of these two quantities follows the model 
#       delta(x,y)  = I0(x,y) - N0(x,y) 
#                   = abs(psi_R(x,y)) * ( abs(psi_R(x,y)) + 2 * sqrt(N0(x,y)) * np.cos( phi0(x,y) - mu ) ) 
#   We know mu (which has no spatial dependence) by design but need to estimate phi0(x,y) and psiR(x,y). 
#   A reasonable prior for phi0(x,y) is a zero mean Guassian with some variance sigma^2_phi. 
#   We also know N0(x,y) is zero outside the pupil (i.e. x,y not in P) - which allows direct sampling of  
#       delta(x,y not in P) =  abs(psi_R(x,y not in P))^2
#   We also have a theoretical model of psi_R(x,y) across all regions that 
#       psi_R(x,y) = A * iFFT( Pi(x,y) * FFT( N0 * exp(1j phi0 (x,y) ) ) where iFFT and FFT are the Fourier transforms, 
#   and A is a known (by construction) parameter and Pi(x,y) defines the phaseshifting region on the mask (0 if no phase shift applie, 1 otherwise).
 

#   The problem is then to estimate phi0(x,y) and psi_R(x,y) from the measurements I0(x,y) and N0(x,y).
#    open question 
#       - fit psi_R first from samples outside pupil and then use this to estimate phi0. 
#     Then repeat using full function of phi0 est, to get psi_R. Is this convex or non-convex? i.e. will it always converge to global maxima?
#       - Assume phi0 = 0 and then solve abs(psi_R) pixelwise. Selecting smoothest solution. 
#          Then use psi_R as normal with some uncertainty and update solve phi0 (bayesian).

# the general pixelwise model:
#           y = a*X^2 + b * X * cos( phi + a )


#%%
#Define the function f(X,ϕ)=a⋅X2+b⋅X⋅cos⁡(ϕ+c) .
# For a fixed value of y, plot the contour where f(X,ϕ)=y


# Given constants
a = 1.0   # Fixed a value

# Define the function f(|psi_R|, phi)
def f(psi_R, phi, psi_A, mu):
    b = 2 * np.sqrt(psi_A)  # Compute b from |psi_A|
    return a * psi_R**2 + b * psi_R * np.cos(phi + mu)

# Define the function for y = |\psi_C|^2 - |\psi_A|^2
def psi_C_squared(psi_R, phi, psi_A, mu):
    return f(psi_R, phi, psi_A, mu) + psi_A**2  # y is relabeled as |\psi_C|^2 - |\psi_A|^2

# Define the range of |psi_R| (X) and phi values
psi_R_vals = np.linspace(-10, 10, 400)  # range of |psi_R| (X) values
phi_vals = np.linspace(-np.pi, np.pi, 400)  # range of phi values

# Create meshgrid for |psi_R| and phi
psi_R, phi = np.meshgrid(psi_R_vals, phi_vals)

# Initial parameters
psi_A_init = 2.0  # Initial |psi_A|
mu_init = 0.5     # Initial mu value
psi_C_init = 10.0  # Initial value for |psi_C|

# Create the figure and the contour plot
fig, ax = plt.subplots(figsize=(8, 6))
plt.subplots_adjust(left=0.1, bottom=0.35)  # Make room for sliders
f_vals = psi_C_squared(psi_R, phi, psi_A_init, mu_init)
contour = ax.contour(psi_R, phi, f_vals, levels=[psi_C_init], colors='r')
ax.set_title(r'Phase Space Contour for $|\psi_C|^2 - |\psi_A|^2 = {}$'.format(psi_C_init))
ax.set_xlabel(r'$|\psi_R|$')
ax.set_ylabel(r'$\phi$')
ax.grid(True)

# Adjust the position of the sliders
ax_slider_psi_A = plt.axes([0.1, 0.20, 0.8, 0.03], facecolor='lightgoldenrodyellow')
ax_slider_mu = plt.axes([0.1, 0.15, 0.8, 0.03], facecolor='lightgoldenrodyellow')
ax_slider_psi_C = plt.axes([0.1, 0.10, 0.8, 0.03], facecolor='lightgoldenrodyellow')

# Define the sliders
slider_psi_A = Slider(ax_slider_psi_A, r'$|\psi_A|$', 0.1, 5.0, valinit=psi_A_init)
slider_mu = Slider(ax_slider_mu, r'$\mu$', -np.pi, np.pi, valinit=mu_init)
slider_psi_C = Slider(ax_slider_psi_C, r'$|\psi_C|^2 - |\psi_A|^2$', 0.1, 20.0, valinit=psi_C_init)

# Update function to redraw the plot when sliders change
def update(val):
    psi_A = slider_psi_A.val
    mu = slider_mu.val
    psi_C = slider_psi_C.val
    
    ax.clear()
    f_vals = psi_C_squared(psi_R, phi, psi_A, mu)
    ax.contour(psi_R, phi, f_vals, levels=[psi_C], colors='r')
    ax.set_title(r'Phase Space Contour for $|\psi_C|^2 - |\psi_A|^2 = {}$'.format(psi_C))
    ax.set_xlabel(r'$|\psi_R|$')
    ax.set_ylabel(r'$\phi$')
    ax.grid(True)
    fig.canvas.draw_idle()

# Attach the update function to sliders
slider_psi_A.on_changed(update)
slider_mu.on_changed(update)
slider_psi_C.on_changed(update)

plt.show()


# I think the best route (similar to what Zelda does) is to estimate psi_R first from measured (clear) pupil
# then reconstruct phase analytically.


#%%

#### AFTER CALIBRATION OF THE OPTICAL GAIN REFERENCE (ZERO ABERRATIONS)
# simulate simulate the ZWFS signal with no-aberrations, es 
# iterate phase screens with vibrations
#   for each phase screen,
#       remove N modes
#       add any modal vibrations
#       calculate the non-observable parameters 
#       - strehl 
#       - optical gain (b)
#       simulate the observable parameters
#       - ZWFS signal
#   goal is to build a model that predicts non-observable parameters from observable parameters
#   baseline model for Strehl is a linear combination of ZWFS intensity in sub-regions of the image  
#   baseline model for the optical gain is then b = sqrt(S) * b0. Where b0 is the optical gain without aberrations  






# initialize our ZWFS instrument
wvl0=1.25e-6
config_ini = 'BaldrApp/baldrapp/configurations/BALDR_UT_J3.ini'
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=wvl0)


# short hand for pupil dimensions (pixels)
#dim = zwfs_ns.grid.N * zwfs_ns.grid.padding_factor # should match zwfs_ns.pyZelda.pupil_dim
# spatial differential in pupil space 
dx = zwfs_ns.grid.D / zwfs_ns.grid.N
# get required simulation sampling rate to match physical parameters 
dt = dx * zwfs_ns.atmosphere.pixels_per_iteration / zwfs_ns.atmosphere.v # s # simulation sampling rate

print(f'current parameters have effective wind velocity = {round(zwfs_ns.atmosphere.v )}m/s')
scrn = ps.PhaseScreenKolmogorov(nx_size=zwfs_ns.grid.dim, pixel_scale=dx, r0=zwfs_ns.atmosphere.r0, L0=zwfs_ns.atmosphere.l0, random_seed=1)


# first stage AO 
basis_cropped = ztools.zernike.zernike_basis(nterms=150, npix=zwfs_ns.pyZelda.pupil_diameter)
# we have padding around telescope pupil (check zwfs_ns.pyZelda.pupil.shape and zwfs_ns.pyZelda.pupil_diameter) 
# so we need to put basis in the same frame  
basis_template = np.zeros( zwfs_ns.pyZelda.pupil.shape )
basis = np.array( [ util.insert_concentric( np.nan_to_num(b, 0), basis_template) for b in basis_cropped] )

pupil_disk = basis[0] # we define a disk pupil without secondary - useful for removing Zernike modes later

Nmodes_removed = 14 # Default will be to remove Zernike modes 

# vibrations 
mode_indicies = [0, 1]
spectrum_type = ['1/f', '1/f']
opd = [50e-9, 50e-9]
vibration_frequencies = [15, 45] #Hz


# calculate reference (perfect system) optical gain (b0)
b0, expi = ztools.create_reference_wave_beyond_pupil(zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate, zwfs_ns.pyZelda.mask_Fratio,
                                       zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wvl0, clear=np.array([]), 
                                       sign_mask=np.array([]), cpix=False)

# to put in pixel space 
b0_pixelspace = bldr.average_subarrays( abs(b0) , (12,12)) 


# input amplitude of the star 
photon_scaling = zwfs_ns.throughput.vlti_throughput * (np.pi * (zwfs_ns.grid.D/2)**2) / (np.pi * zwfs_ns.pyZelda.pupil_diameter/2)**2 * util.magnitude_to_photon_flux(magnitude=zwfs_ns.stellar.magnitude, band = zwfs_ns.stellar.waveband, wavelength= 1e9*wvl0)

telemetry = {
    'I0':[],
    'N0':[],
    'scrn':[],
    'ao_1':[],
    'Ic':[],
    'i':[],
    'strehl':[],
    'dm_cmd':[],
    'b':[],
    'b_pixelspace':[],
    'ao_2':[]
}

telem_ns = SimpleNamespace(**telemetry)

it_grid = np.arange(0, 100)
scrn_scaling_grid = np.logspace(0,1,10)
dm_phasescreen = True # apply the phasescreen to the DM (this is how we will calibrate the real system)
# DO THIS - BUT PUT ABERRATION ON THE DM!!!! (SAME WAY WE WOULD CALIBRATE THE REAL SYSTEM)

scrn_list = []
for _ in range(100):
    #scrn = ps.PhaseScreenKolmogorov(nx_size=zwfs_ns.grid.N, pixel_scale=dx, r0=zwfs_ns.atmosphere.r0, L0=zwfs_ns.atmosphere.l0, random_seed=1)
    scrn = ps.PhaseScreenKolmogorov(nx_size=24, pixel_scale = zwfs_ns.grid.D / 24, r0=zwfs_ns.atmosphere.r0, L0=zwfs_ns.atmosphere.l0, random_seed=None)
    scrn_list.append( scrn ) 
    #zwfs_ns.grid.pupil_mask * util.insert_concentric( scrn.scrn, zwfs_ns.pyZelda.pupil ) )

for it in range(len(scrn_list)):

    # roll screen
    #scrn.add_row()     
    for ph_scale in scrn_scaling_grid: 
        

        if dm_phasescreen:
            #scaling_factor=0.05, drop_indicies = [0, 11, 11 * 12, -1] , plot_cmd=False
            zwfs_ns.dm.current_cmd =  util.create_phase_screen_cmd_for_DM(scrn_list[it],  scaling_factor=ph_scale , drop_indicies = [0, 11, 11 * 12, -1] , plot_cmd=False) 
        
            opd_current_dm = bldr.get_dm_displacement( command_vector= zwfs_ns.dm.current_cmd   , gain=zwfs_ns.dm.opd_per_cmd, \
                sigma= zwfs_ns.grid.dm_coord.act_sigma_wavesp, X=zwfs_ns.grid.wave_coord.X, Y=zwfs_ns.grid.wave_coord.Y,\
                    x0=zwfs_ns.grid.dm_coord.act_x0_list_wavesp, y0=zwfs_ns.grid.dm_coord.act_y0_list_wavesp )
            
            phi = zwfs_ns.grid.pupil_mask  *  2*np.pi / zwfs_ns.optics.wvl0 * (  opd_current_dm  )
            
            pupil_disk_cropped, atm_in_pupil = util.crop_pupil(pupil_disk, phi)
        else: 
            # crop the pupil disk and the phasescreen within it (remove padding outside pupil)
            pupil_disk_cropped, atm_in_pupil = util.crop_pupil(pupil_disk, ph_scale * scrn.scrn)
            
            
        # test project onto Zernike modes 
        mode_coefficients = np.array( ztools.zernike.opd_expand(atm_in_pupil * pupil_disk_cropped,\
            nterms=len(basis), aperture =pupil_disk_cropped))

        # do the reconstruction for N modes
        reco = np.sum( mode_coefficients[:Nmodes_removed,np.newaxis, np.newaxis] * basis[:Nmodes_removed,:,:] ,axis = 0) 

        # remove N modes 
        
        if dm_phasescreen:
            ao_1 = zwfs_ns.atmosphere.scrn_scaling * pupil_disk * (phi - reco) 
        else:
            ao_1 = zwfs_ns.atmosphere.scrn_scaling * pupil_disk * (scrn.scrn - reco)     
            
        # add vibrations
        # TO DO 

        # for calibration purposes
        print( f'for {Nmodes_removed} Zernike modes removed (scrn_scaling={ph_scale}),\n \
            atmospheric conditions r0= {round(zwfs_ns.atmosphere.r0,2)}m at a central wavelength {round(1e6*wvl0,2)}um\n\
                post 1st stage AO rmse [nm rms] = ',\
            round( 1e9 * (wvl0 / (2*np.pi) * ao_1)[zwfs_ns.pyZelda.pupil>0.5].std() ) )


        # apply DM 
        #ao1 *= DM_field

        # convert to OPD map
        opd_map = zwfs_ns.pyZelda.pupil * wvl0 / (2*np.pi) * ao_1 
        
        if it==0:
            N0 = ztools.propagate_opd_map(0*opd_map, zwfs_ns.pyZelda.mask_diameter, 0*zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate,
                                            zwfs_ns.pyZelda.mask_Fratio, zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wave=wvl0)

            I0 = ztools.propagate_opd_map(0*opd_map, zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate,
                                            zwfs_ns.pyZelda.mask_Fratio, zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wave=wvl0)
        
            telem_ns.I0.append(I0)
            telem_ns.N0.append(N0)  

        # caclulate Strehl ratio
        strehl = np.exp( - np.var( ao_1[zwfs_ns.pyZelda.pupil>0.5]) )

        b, _ = ztools.create_reference_wave_beyond_pupil_with_aberrations(opd_map, zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate, zwfs_ns.pyZelda.mask_Fratio,
                                            zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wvl0, clear=np.array([]), 
                                            sign_mask=np.array([]), cpix=False)


        b_pixelspace = bldr.average_subarrays( abs(b) , (zwfs_ns.detector.binning, zwfs_ns.detector.binning)) 

        # normalized such that np.sum( I0 ) / np.sum( N0 ) ~ 1 where N0.max() = 1. 
        # do normalization by known area of the pupil and the input stellar magnitude at the given wavelength 
        # represent as #photons / s / pixel / nm

        Ic = photon_scaling * zwfs_ns.pyZelda.propagate_opd_map( opd_map , wave = wvl0 )

        det_binning = round( bldr.calculate_detector_binning_factor(grid_pixels_across_pupil = zwfs_ns.pyZelda.pupil_diameter, detector_pixels_across_pupil = 12) )

        i = bldr.detect( Ic, binning = (zwfs_ns.detector.binning, zwfs_ns.detector.binning), qe=zwfs_ns.detector.qe , dit=zwfs_ns.detector.dit, ron= zwfs_ns.detector.ron, include_shotnoise=True, spectral_bandwidth = zwfs_ns.stellar.bandwidth )

        #telem_ns.ao_1.append(zwfs_ns.pyZelda.pupil * ao_1)
        telem_ns.i.append(i)
        telem_ns.Ic.append(Ic)
        telem_ns.strehl.append(strehl)
        telem_ns.b.append(b)
        telem_ns.b_pixelspace.append(b_pixelspace)
        telem_ns.dm_cmd.append(zwfs_ns.dm.current_cmd )
        
    print( f'iteration {it} done')

# IF YOU WANT TO VISUALIZE ANY INTERMEDIATE STEPS
#plt.figure(); plt.imshow( zwfs_ns.pyZelda.pupil * scrn.scrn); plt.show()
#plt.imshow( reco ); plt.colorbar(); plt.show()
#plt.imshow( ao_1 ); plt.colorbar(); plt.show()
#plt.imshow( zwfs_ns.pyZelda.pupil * np.abs(b0) ); plt.colorbar(); plt.show()
#plt.imshow( zwfs_ns.pyZelda.pupil * np.angle(b0) ); plt.colorbar(); plt.show()
#plt.imshow( zwfs_ns.pyZelda.pupil * np.abs(b) ); plt.colorbar(); plt.show()
#plt.imshow( zwfs_ns.pyZelda.pupil * np.angle(b) ); plt.colorbar(); plt.show()
#plt.imshow( Ic ); plt.show()
#plt.imshow( i ) ; plt.colorbar(); plt.show()

# Example usage:
# example_images = [telem_ns.ao_1, telem_ns.strehl, telem_ns.i, [abs(b) for b in telem_ns.b] ]  # 3 lists of 10 images each

# # Call the function to dynamically slide through the images
# util.display_images_with_slider(example_images)
# # Call the function to create and save the movie
# util.display_images_as_movie(example_images, plot_titles=None, cbar_labels=None, save_path="output_movie.mp4", fps=5)

"""
observable is DM command and ZWFS signal

non-observable is Strehl and optical gain that we wish to model 

use known DM commands to calibrate Strehl and optical gain model 

we want to model optical gain in the pixel space. 
Therefore any iteration model can be scaled by it.

proceedure: 

roll various phase screens across DM and get ZWFS intensity response 

dm shape related to strehl via DM influence function

filter pixels that above Pearson correlation threshold
fit a linear model to the data ( S ~ exp(-var(dm_rmse)) ~ sum_i alpha_i * I_i + i0_i )

verify optical gain model 
b = sqrt(S) * b0

metrics 
- rmse inside the pupil vs strehl

for given strehl bin  
- mean residual vs radial profile (look at bias in the model)
- rmse vs radial profile 


store the results in a config file 
- strehl model pixels  
- coefficients
- b0 

"""

fig_path = '/home/benja/Downloads/'
# Example usage
correlation_map = compute_correlation_map(np.array( telem_ns.i ), np.array( telem_ns.strehl) )

# SNR 
SNR = np.mean( telem_ns.i ,axis=0 ) / np.std( telem_ns.i ,axis=0  )

im_list =  [ correlation_map ]
cbar_list = ['pearson R']

util.nice_heatmap_subplots( im_list = [ correlation_map ] , cbar_label_list = ['Pearson R'] , \
    savefig = None) #fig_path + 'strehl_vs_intensity_pearson_R.png' )

util.nice_heatmap_subplots( im_list = [ SNR / np.max( SNR ) ] , cbar_label_list = ['normalized SNR'] ,\
    savefig = None)# fig_path + 'SNR_simulation.png' )

# Select top 5% of pixels with the highest correlation
threshold = 0.9
selected_pixels = correlation_map > threshold

plt.figure()
plt.imshow( selected_pixels)
plt.colorbar(label = "filter")
plt.savefig(fig_path + 'selected_pixels.png', bbox_inches='tight', dpi=300)
plt.show()

## FITTING THE MODEL 
model_description = "Linear regression model fitting intensities to Strehl ratio."

model = StrehlModel(model_description)

#pixel_indices = np.where( selected_pixels )

i_train = int( 0.6 * len( telem_ns.i ) )

y_train = np.array(  telem_ns.strehl )[:i_train]
X_train = np.array( telem_ns.i )[:i_train] 

y_test = np.array(  telem_ns.strehl )[i_train:]
X_test = np.array( telem_ns.i )[i_train:]


#coefficients, intercept = model.fit_linear_model(x, y)
model.fit(X = X_train,\
        y = y_train ,\
        pixel_filter=selected_pixels )

y_fit = model.apply_model(X_test) 

# add the pupil in 
model.name = zwfs_ns.name # so we know what config file was used 

#y_fit = model.predict(x)

# show out of sample test results 
util.plot_data_and_residuals(y_test, y_test, y_fit, xlabel=r'$\text{Strehl Ratio}$', ylabel=r'$\text{Predicted Strehl Ratio}$', \
    residual_ylabel=r'$\Delta$',label_1="1:1", label_2="model", savefig='{}strehl_linear_fit.png'.format(fig_path) )


# save the model
model.save_model_to_pickle(filename=fig_path + 'strehl_model.pkl')

# read it in 
model = load_model_from_pickle(filename=fig_path + 'strehl_model.pkl')



### MODELLING OPTICAL GAIN - VERIFICATION 
"""
# baseline model is b = sqrt(S) * b0
# quality metric = rmse inside the pupil! 
# using meeasured Strehl (which we won't have in the real system)
"""

strehl = np.array( telem_ns.strehl )
y = np.array( [abs(b)for b in  telem_ns.b ] )

# using real strehl values
#y_model = np.array( [ss* np.abs(b0) for ss in np.sqrt( np.array( telem_ns.strehl ) ) ] )

# using the strehl model 
y_model = np.array( [ss * np.abs(b0) for ss in  np.sqrt( model.apply_model( np.array( telem_ns.i  ) ) ) ] )

## get azimuth statistics 
#y_az = np.array([imutils.profile(yy, ptype='mean', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)[0] for yy in y])


# to adjust for small statistics we use Bessel correction for variance
img = y[0].copy()

# array dimensions
dimx = img.shape[1]
dimy = img.shape[0]
# center
center = (dimx // 2, dimy // 2)

# intermediate cartesian arrays
xx, yy = np.meshgrid(np.arange(dimx, dtype=np.int64) - center[0], np.arange(dimy, dtype=np.int64) - center[1])
rr = np.sqrt(xx**2 + yy**2)

# rounds for faster calculation
rr = np.round(rr, decimals=0)

# find unique radial values
uniq = np.unique(rr, return_inverse=True, return_counts=True)

r_uniq_val = uniq[0]
r_uniq_inv = uniq[1]
r_uniq_cnt = uniq[2]


# get radial vector in pixels 
_, r_pix  = imutils.profile(y[0], ptype='mean', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)

mean_b0_az = imutils.profile(abs(b0), ptype='mean', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)[0] 
mean_residual_az = np.array([imutils.profile(meas-model, ptype='mean', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)[0] for meas,model in zip(y, y_model) ])
std_residual_az = np.array([imutils.profile(meas-model, ptype='std', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)[0] for meas,model in zip(y, y_model) ])
# editted function for ptype = var using Bessel correction for small sample size 
var_residual_az = np.array([imutils.profile(meas-model, ptype='var', step=1, mask=None, center=None, rmax=0, clip=True, exact=False)[0] for meas,model in zip(y, y_model) ])

bessel_correction = np.array( [rr / (rr - 1) if rr>1 else np.nan for rr in r_uniq_cnt] )

# Normalize r_pix
r_pix_normalized = r_pix / (zwfs_ns.grid.N / 2)

# Define Strehl bins and labels
strehl_bins = [(0.4, 0.6), (0.6, 0.8), (0.8, 1.0)]
colors = ['blue', 'green', 'red']
labels = ['0.4 ≤ Strehl < 0.6', '0.6 ≤ Strehl < 0.8', '0.8 ≤ Strehl < 1.0']

# Create the plot
plt.figure(figsize=(8, 6))

for (low, high), color, label in zip(strehl_bins, colors, labels):
    # Define the Strehl filter for the current bin
    strehl_filter = (strehl >= low) & (strehl < high)
    
    # Calculate mean and standard deviation for the current Strehl bin
    mean_residual = np.mean(mean_residual_az[strehl_filter], axis=0)
    #std_residual = np.mean( std_residual_az[strehl_filter], axis=0)  # Standard deviation
    #std_residual = np.mean( np.sqrt(var_residual_az[strehl_filter] ), axis=0)  # editted function for ptype = var using Bessel correction for small sample size 
    std_residual = np.std( mean_residual_az[strehl_filter], axis=0, ddof=1)  # Standard deviation
    
    # Plot the mean as a solid line for the current bin
    plt.plot(r_pix_normalized, mean_residual/mean_b0_az, color=color, label=label)
    
    # Fill the area around the mean within 1 standard deviation
    plt.fill_between(r_pix_normalized, (mean_residual - std_residual)/mean_b0_az, 
                     (mean_residual + std_residual)/mean_b0_az, color=color, alpha=0.2)

# Formatting the plot
plt.ylabel('Expected Fractional Residual\n'+r'$\left \langle \frac{|b| - |b_{model}|}{b_0} \right \rangle$',fontsize=15)
plt.xlabel(r'Fractional Pupil Radius',fontsize=15)
plt.xlim( [0, 1.05])
plt.gca().tick_params(axis='both', which='major', labelsize=15)
#plt.title('Expected Residual and Standard Deviation for Different Strehl Bins')
plt.grid(True)
plt.legend(loc='best',fontsize=15)  # Add a legend to differentiate Strehl bins


# Show the plot
plt.tight_layout()
plt.savefig(fig_path + 'residuals_optical_gain_radial_profile_WITH_STREHL_MODEL.png', bbox_inches='tight', dpi=300)
plt.show()







import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.interpolate import RectBivariateSpline
import glob 
from astropy.io import fits 
import os 
import json
import datetime 
from scipy.interpolate import griddata
from scipy.ndimage import map_coordinates

# Function to get indices for the inner square on DM, accounting for missing corners
def get_inner_square_indices(outer_size, inner_offset, without_outer_corners=True):
    
    if without_outer_corners: # assumes outer corners of DM are missing 
        
        # Create a 12x12 grid with missing corners marked by NaN
        grid = np.arange(outer_size**2).reshape(outer_size, outer_size).astype(float)
        
        # Mark the missing corners
        grid[0, 0] = grid[0, -1] = grid[-1, 0] = grid[-1, -1] = np.nan

        # Define the inner square boundaries
        top_row = inner_offset
        bottom_row = outer_size - inner_offset - 1
        left_col = inner_offset
        right_col = outer_size - inner_offset - 1

        # Get the indices of the four corners of the inner square
        inner_corners = [
            grid[top_row, left_col],
            grid[top_row, right_col],
            grid[bottom_row, left_col],
            grid[bottom_row, right_col]
        ]

        # Remove the NaNs from the grid (we only want valid actuator indices)
        valid_indices = np.isfinite(grid.flatten())
        valid_actuators = np.where(valid_indices)[0]

        # Convert the inner square corner indices to valid actuator indices
        inner_square_indices = [np.where(valid_actuators == np.where(grid.flatten() == corner)[0][0])[0][0] for corner in inner_corners]
    else:  # outer corners not missing - this case is important with pinned edge actuators since the modal space then does not have missing corners! 
        grid = np.arange(outer_size**2).reshape(outer_size, outer_size)

        # Define the inner square boundaries
        top_row = inner_offset
        bottom_row = outer_size - inner_offset - 1
        left_col = inner_offset
        right_col = outer_size - inner_offset - 1

        # Get the indices of the four corners of the inner square
        inner_corners = [
            grid[top_row, left_col],
            grid[top_row, right_col],
            grid[bottom_row, left_col],
            grid[bottom_row, right_col]
        ]

        # Convert the inner square corner indices to valid actuator indices
        inner_square_indices = [np.where(grid.flatten() == corner)[0][0] for corner in inner_corners]

    return inner_square_indices



# Define a 2D Gaussian function
def gaussian_2d(xy, amp, x0, y0, sigma_x, sigma_y, theta, offset):
    x, y = xy
    x0, y0 = float(x0), float(y0)
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    
    return offset + amp * np.exp(-(a*(x-x0)**2 + 2*b*(x-x0)*(y-y0) + c*(y-y0)**2))


# Function to get adjacent pixels within a given radius
def get_adjacent_pixels(coord, radius, pixel_values):
    """
    I, J: Coordinates of the central pixel
    radius: Radius (in pixels) for the region of interest
    pixel_values: 2D array of the image pixel values
    
    Returns:
    - Adjacent pixel indices and values within the radius
    """
    n,m = coord
    y, x = np.ogrid[-radius:radius+1, -radius:radius+1]
    mask = x**2 + y**2 <= radius**2
    indices = np.argwhere(mask)
    indices[:, 0] += n - radius
    indices[:, 1] += m - radius
    valid_indices = [(i, j) for i, j in indices if i >= 0 and j >= 0 and i < pixel_values.shape[0] and j < pixel_values.shape[1]]
    pixel_values_adj = np.array([pixel_values[i, j] for i, j in valid_indices])
    return np.array(valid_indices), pixel_values_adj


# Spline interpolation and Gaussian fitting
def interpolate_and_fit_gaussian(coord, radius, pixel_values, factor=5):
    """
    coord = tuple of coordinates of the central pixel (row, column)
    radius: Radius of the neighborhood
    pixel_values: 2D array of pixel values (image)
    factor: Upsampling factor for spline interpolation (default 5x finer grid)
    
    Returns:
    A dictionary with fitted Gaussian parameters and residuals.
    """
    i, j = coord
    # Get the adjacent pixels within the radius
    indices, pixel_values_adj = get_adjacent_pixels((i, j), radius, pixel_values)
    x = indices[:, 1]
    y = indices[:, 0]
    
    # Create a coarse grid
    x_grid = np.unique(x)
    y_grid = np.unique(y)
    X, Y = np.meshgrid(x_grid, y_grid)
    
    # Get pixel values on the coarse grid
    Z = pixel_values[np.ix_(y_grid, x_grid)]
    
    # Spline interpolate onto a finer grid
    spline = RectBivariateSpline(y_grid, x_grid, Z)
    
    # Create a finer grid
    x_fine = np.linspace(x_grid.min(), x_grid.max(), factor * len(x_grid))
    y_fine = np.linspace(y_grid.min(), y_grid.max(), factor * len(y_grid))
    X_fine, Y_fine = np.meshgrid(x_fine, y_fine)
    Z_fine = spline(y_fine, x_fine)
    
    # Initial guess for Gaussian fitting
    initial_guess = (Z_fine.max(), np.mean(x_fine), np.mean(y_fine), 1, 1, 0, np.median(Z_fine))
    
    # Perform Gaussian fitting
    try:
        popt, pcov = curve_fit(gaussian_2d, (X_fine.ravel(), Y_fine.ravel()), Z_fine.ravel(), p0=initial_guess)
    except RuntimeError:
        return {"error": "Fit did not converge"}
    
    # Calculate residuals
    Z_fit = gaussian_2d((X_fine, Y_fine), *popt).reshape(X_fine.shape)
    residuals = Z_fine - Z_fit
    
    # Extract fitted parameters
    amp_fit, x0_fit, y0_fit, sigma_x_fit, sigma_y_fit, theta_fit, offset_fit = popt
    
    # Convert the fine-grid coordinates of the peak to the global pixel frame
    x0_global = x0_fit
    y0_global = y0_fit
    
    # Create a dictionary to store the results
    fit_dict = {
        "amplitude": amp_fit,
        "x0": x0_global,
        "y0": y0_global,
        "sigma_x": sigma_x_fit,
        "sigma_y": sigma_y_fit,
        "theta": theta_fit,
        "offset": offset_fit,
        "X":X,
        "Y":Y,
        "I_raw":Z,
        "X_interp":X_fine,
        "Y_interp":Y_fine,
        "I_interp":Z_fine,
        "I_fit": Z_fit,
        "residuals": residuals,
         
    }
    
    return fit_dict



def plot_fit_results(fit_dict, savefig = None):
    """
    Plots the original, interpolated, fitted, and residual images from the fit_dict, with X and Y axes.
    
    Args:
        fit_dict: Dictionary returned by the interpolate_and_fit_gaussian function.
    """
    # Extract data from the fit_dict
    I = fit_dict["I_raw"]  # Original image
    I_interp = fit_dict["I_interp"]  # Interpolated image
    I_fit = fit_dict["I_fit"]  # Fitted Gaussian model
    residuals = fit_dict["residuals"]  # Residuals

    # Extract the grid coordinates
    X = fit_dict["X"]
    Y = fit_dict["Y"]
    X_interp = fit_dict["X_interp"]
    Y_interp = fit_dict["Y_interp"]

    # Create a figure with 4 subplots (in a single row)
    fig, axes = plt.subplots(1, 4, figsize=(16, 4))

    # Plot Original Image (I) with X, Y axis
    im1 = axes[0].imshow(I, origin='lower', extent=[X.min(), X.max(), Y.min(), Y.max()], cmap='viridis')
    axes[0].set_title('Original Image')
    #axes[0].set_xlabel('X')
    #axes[0].set_ylabel('Y')
    cbar1 = plt.colorbar(im1, ax=axes[0], orientation='horizontal', pad=0.1)
    cbar1.set_label('Intensity')

    # Plot Interpolated Image (I_interp) with X_interp, Y_interp axis
    im2 = axes[1].imshow(I_interp, origin='lower', extent=[X_interp.min(), X_interp.max(), Y_interp.min(), Y_interp.max()], cmap='viridis')
    axes[1].set_title('Interpolated Image')
    #axes[1].set_xlabel('X_interp')
    #axes[1].set_ylabel('Y_interp')
    cbar2 = plt.colorbar(im2, ax=axes[1], orientation='horizontal', pad=0.1)
    cbar2.set_label('Interpolated Intensity')

    # Plot Fitted Gaussian Model (I_fit) with X_interp, Y_interp axis
    im3 = axes[2].imshow(I_fit, origin='lower', extent=[X_interp.min(), X_interp.max(), Y_interp.min(), Y_interp.max()], cmap='viridis')
    axes[2].set_title('Fitted Gaussian')
    #axes[2].set_xlabel('X_interp')
    #axes[2].set_ylabel('Y_interp')
    cbar3 = plt.colorbar(im3, ax=axes[2], orientation='horizontal', pad=0.1)
    cbar3.set_label('Fitted Intensity')

    # Plot Residuals with X_interp, Y_interp axis
    im4 = axes[3].imshow(residuals, origin='lower', extent=[X_interp.min(), X_interp.max(), Y_interp.min(), Y_interp.max()], cmap='coolwarm')
    axes[3].set_title('Residuals')
    #axes[3].set_xlabel('X_interp')
    #axes[3].set_ylabel('Y_interp')
    cbar4 = plt.colorbar(im4, ax=axes[3], orientation='horizontal', pad=0.1)
    cbar4.set_label('Residual Intensity')

    # Display the plots
    plt.tight_layout()
    if savefig :
        plt.savefig(savefig) 
    plt.show()


# Function to calculate the intersection of two lines
def find_intersection(x1, y1, x2, y2, x3, y3, x4, y4):
    """
    Find the intersection point of two lines:
    Line 1: passing through (x1, y1) and (x2, y2)
    Line 2: passing through (x3, y3) and (x4, y4)
    """

    # Line 1 coefficients (a1*x + b1*y = c1)
    a1 = y2 - y1
    b1 = x1 - x2
    c1 = a1 * x1 + b1 * y1

    # Line 2 coefficients (a2*x + b2*y = c2)
    a2 = y4 - y3
    b2 = x3 - x4
    c2 = a2 * x3 + b2 * y3

    # Calculate the determinant of the system
    det = a1 * b2 - a2 * b1

    if det == 0:
        raise ValueError("The lines are parallel and do not intersect.")

    # Calculate the intersection point
    x_intersect = (b2 * c1 - b1 * c2) / det
    y_intersect = (a1 * c2 - a2 * c1) / det

    return x_intersect, y_intersect


# def sort_corners(corners):
#     # THIS FUNCTION FAILS WHEN CORNERS ARE HORIZONTALLY OR VERTICALLY ALIGNED 
#     sorted_corners = sorted(corners, key=lambda p: (p[0], p[1]))

#     # Separate them into logical quadrilateral corners
#     top_left = sorted_corners[0]
#     top_right = sorted_corners[1] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[2]
#     bottom_left = sorted_corners[2] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[1]
#     bottom_right = sorted_corners[3]
    
#     return(top_left, top_right,  bottom_left , bottom_right )

def sort_corners(corners):
    # Convert the corners to a numpy array
    corners = np.array(corners)

    # Calculate the centroid (geometric center) of the four points
    centroid = np.mean(corners, axis=0)

    # Sort the points based on their angle with respect to the centroid
    def angle_from_centroid(point):
        return np.arctan2(point[1] - centroid[1], point[0] - centroid[0])

    sorted_corners = sorted(corners, key=angle_from_centroid)

    # Assign corners: the sorting will order the points in a consistent counter-clockwise order
    # Now label them accordingly
    top_left = sorted_corners[3]
    top_right = sorted_corners[2]
    bottom_right = sorted_corners[1]
    bottom_left = sorted_corners[0]

    return top_left, top_right, bottom_left, bottom_right

# Main function to compute the intersection of diagonals of a quadrilateral
def find_quadrilateral_diagonal_intersection(corners ):
    """
    Given four corners of a quadrilateral, find the intersection point of its diagonals.
    
    Args:
    - corners: A list of 4 tuples, each representing the (x, y) coordinates of a corner.
    
    Output:
    - Intersection point (x, y) of the diagonals
    """
    # Sort the corners by x value first, then by y value for easier identification of top-left, top-right, etc.
    #sorted_corners = sorted(corners, key=lambda p: (p[0], p[1]))

    # Separate them into logical quadrilateral corners
    #top_left = sorted_corners[0]
    #top_right = sorted_corners[1] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[2]
    #bottom_left = sorted_corners[2] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[1]
    #bottom_right = sorted_corners[3]

    top_left, top_right,  bottom_left , bottom_right  = sort_corners(corners)
    # Diagonal 1: passing through top-left and bottom-right
    # Diagonal 2: passing through top-right and bottom-left
    intersection = find_intersection(top_left[0], top_left[1], bottom_right[0], bottom_right[1], 
                                     top_right[0], top_right[1], bottom_left[0], bottom_left[1])
    
    return intersection #, [top_left, top_right, bottom_right, bottom_left]


# Plotting function to visualize the quadrilateral and diagonals
def plot_quadrilateral_with_diagonals(corners, intersection):
    # Unpack the corners
    #sorted_corners = sorted(corners, key=lambda p: (p[0], p[1]))

    # Separate them into logical quadrilateral corners
    #top_left = sorted_corners[0]
    #top_right = sorted_corners[1] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[2]
    #bottom_left = sorted_corners[2] if sorted_corners[1][1] > sorted_corners[0][1] else sorted_corners[1]
    #bottom_right = sorted_corners[3]

    top_left, top_right,  bottom_left , bottom_right  = sort_corners(corners) 
    # Plot the quadrilateral
    plt.figure(figsize=(6, 6))
    
    # Plot the quadrilateral edges
    x_vals = [top_left[0], top_right[0], bottom_right[0], bottom_left[0], top_left[0]]
    y_vals = [top_left[1], top_right[1], bottom_right[1], bottom_left[1], top_left[1]]
    plt.plot(x_vals, y_vals, 'b-', label='Quadrilateral')

    # Plot the diagonals
    plt.plot([top_left[0], bottom_right[0]], [top_left[1], bottom_right[1]], 'r--', label='Diagonal 1')
    plt.plot([top_right[0], bottom_left[0]], [top_right[1], bottom_left[1]], 'g--', label='Diagonal 2')

    # Plot the intersection point
    plt.plot(intersection[0], intersection[1], 'ko', label='Intersection', markersize=10)

    # Label the points
    plt.text(top_left[0], top_left[1], 'Top Left', fontsize=12, verticalalignment='bottom')
    plt.text(top_right[0], top_right[1], 'Top Right', fontsize=12, verticalalignment='bottom')
    plt.text(bottom_left[0], bottom_left[1], 'Bottom Left', fontsize=12, verticalalignment='top')
    plt.text(bottom_right[0], bottom_right[1], 'Bottom Right', fontsize=12, verticalalignment='top')

    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Quadrilateral with Diagonals and Intersection')
    plt.grid(True)
    plt.legend()
    plt.show()



def get_DM_command_in_2D(cmd,Nx_act=12):
    # function so we can easily plot the DM shape (since DM grid is not perfectly square raw cmds can not be plotted in 2D immediately )
    #puts nan values in cmd positions that don't correspond to actuator on a square grid until cmd length is square number (12x12 for BMC multi-2.5 DM) so can be reshaped to 2D array to see what the command looks like on the DM.
    corner_indices = [0, Nx_act-1, Nx_act * (Nx_act-1), Nx_act*Nx_act]
    cmd_in_2D = list(cmd.copy())
    for i in corner_indices:
        cmd_in_2D.insert(i,np.nan)
    return( np.array(cmd_in_2D).reshape(Nx_act,Nx_act) )



def generate_dm_coordinates(Nx=12, Ny=12, spacing=1.0):
    """
    Generates the x, y coordinates of the actuators in a 12x12 grid DM with missing corners.
    
    Args:
        Nx, Ny: Number of actuators in the x and y directions (12x12 grid).
        spacing: The spacing between actuators (default is 1 unit).
    
    Returns:
        - coords: A list of tuples (x, y) representing the coordinates of the actuators.
        - flattened_indices: A dictionary that maps actuator indices (0 to 139) to (x, y) coordinates.
        - coord_to_index: A dictionary mapping (x, y) coordinates to actuator indices.
    """
    coords = []
    coord_to_index = {}
    flattened_indices = {}
    
    center_x = (Nx - 1) / 2  # Center of the grid in x
    center_y = (Ny - 1) / 2  # Center of the grid in y
    
    actuator_index = 0
    for i in range(Ny):
        for j in range(Nx):
            # Skip the missing corners
            if (i == 0 and j == 0) or (i == 0 and j == Nx - 1) or (i == Ny - 1 and j == 0) or (i == Ny - 1 and j == Nx - 1):
                continue

            # Calculate x and y coordinates relative to the center
            x = (j - center_x) * spacing
            y = (i - center_y) * spacing
            
            coords.append((x, y))
            coord_to_index[(x, y)] = actuator_index
            flattened_indices[actuator_index] = (x, y)
            actuator_index += 1

    return coords, flattened_indices, coord_to_index


def get_nearest_actuator(x, y, flattened_indices):
    """
    Finds the nearest actuator index for a given (x, y) coordinate.
    
    Args:
        x, y: The (x, y) coordinates to match to the nearest actuator.
        flattened_indices: A dictionary mapping actuator indices to (x, y) coordinates.
    
    Returns:
        Nearest actuator index.
    """
    distances = {index: np.sqrt((x - coord[0])**2 + (y - coord[1])**2) for index, coord in flattened_indices.items()}
    return min(distances, key=distances.get)


def actuator_to_xy(actuator_index, flattened_indices):
    """
    Given an actuator index, return the corresponding (x, y) coordinates.
    
    Args:
        actuator_index: The actuator number in the flattened 140-length array.
        flattened_indices: A dictionary mapping actuator indices to (x, y) coordinates.
    
    Returns:
        (x, y) coordinates of the actuator.
    """
    return flattened_indices.get(actuator_index)


def fit_affine_transformation_with_center(corners_dm, corners_img, intersection_img):
    """
    Fit an affine transformation from DM space to image space, using the DM center as the origin (0,0).
    
    Args:
        corners_dm: List of (x, y) coordinates of DM corners in DM space (relative to the DM center).
        corners_img: List of (x, y) coordinates of the corresponding points in image space.
        intersection_img: The (x, y) coordinates of the DM center in image space.
    
    Returns:
        - transform_matrix: A 2x3 matrix that transforms DM coordinates to pixel coordinates.
    """
    # Create arrays for the corners
    dm = np.array(corners_dm)
    img = np.array(corners_img)

    # Subtract the DM center (intersection) from the image coordinates to compute translation
    tx, ty = intersection_img
    
    # Now we need to solve for the linear transformation matrix (a, b, c, d)
    # We have the relationship: [x_img, y_img] = A * [x_dm, y_dm] + [tx, ty]
    # where A is the 2x2 matrix with components [a, b; c, d]
    
    # Create the matrix for DM space (without the translation part)
    dm_coords = np.vstack([dm.T, np.ones(len(dm))]).T
    
    # Subtract translation from image coordinates (image coordinates relative to DM center)
    img_coords = img - np.array([tx, ty])

    # Solve the linear system A * dm_coords = img_coords for A (a, b, c, d)
    # Solve the two systems independently for x and y
    A_x = np.linalg.lstsq(dm_coords[:, :2], img_coords[:, 0], rcond=None)[0]
    A_y = np.linalg.lstsq(dm_coords[:, :2], img_coords[:, 1], rcond=None)[0]
    
    # Construct the 2x3 affine transformation matrix
    transform_matrix = np.array([
        [A_x[0], A_x[1], tx],  # [a, b, tx]
        [A_y[0], A_y[1], ty]   # [c, d, ty]
    ])
    
    return transform_matrix

def pixel_to_dm(pixel_coord, transform_matrix):
    """
    Converts pixel coordinates to DM coordinates using the inverse of the affine transformation.
    
    Args:
        pixel_coord: A tuple (x, y) in pixel space.
        transform_matrix: The affine transformation matrix from DM space to pixel space.
    
    Returns:
        Tuple (x_dm, y_dm) in DM coordinates.
    """
    A = transform_matrix[:, :2]  # 2x2 matrix part
    t = transform_matrix[:, 2]   # translation part
    
    # Inverse transformation
    A_inv = np.linalg.inv(A)
    pixel_coord = np.array(pixel_coord)
    dm_coord = np.dot(A_inv, pixel_coord - t)
    return tuple(dm_coord)

def dm_to_pixel(dm_coord, transform_matrix):
    """
    Converts DM coordinates to pixel coordinates using the affine transformation.
    
    Args:
        dm_coord: A tuple (x, y) in DM space.
        transform_matrix: The affine transformation matrix from DM space to pixel space.
    
    Returns:
        Tuple (x_pixel, y_pixel) in pixel coordinates.
    """
    dm_coord = np.array(dm_coord)
    pixel_coord = np.dot(transform_matrix[:, :2], dm_coord) + transform_matrix[:, 2]
    return tuple(pixel_coord)



def convert_to_serializable(obj):
    """
    Recursively converts NumPy arrays and other non-serializable objects to serializable forms.
    Also converts dictionary keys to standard types (str, int, float).
    """
    if isinstance(obj, np.ndarray):
        return obj.tolist()  # Convert NumPy arrays to lists
    elif isinstance(obj, np.integer):
        return int(obj)  # Convert NumPy integers to Python int
    elif isinstance(obj, np.floating):
        return float(obj)  # Convert NumPy floats to Python float
    elif isinstance(obj, dict):
        return {str(key): convert_to_serializable(value) for key, value in obj.items()}  # Ensure keys are strings
    elif isinstance(obj, list):
        return [convert_to_serializable(item) for item in obj]
    else:
        return obj  # Base case: return the object itself if it doesn't need conversion


# def interpolate_pixel_intensities(image, pixel_coords):
#     """
#     Interpolates pixel intensities from an image onto the specified actuator pixel coordinates.
    
#     Args:
#         image: 2D array of pixel intensities (image).
#         pixel_coords: 2D array of actuator coordinates in pixel space (from transform_dict['actuator_coord_list_pixel_space']).
        
#     Returns:
#         Interpolated intensities at the given actuator pixel coordinates.
#     """
#     # Create a grid of original pixel coordinates
#     y, x = np.mgrid[0:image.shape[0], 0:image.shape[1]]
    
#     # Flatten the image and grid for interpolation
#     points = np.vstack((x.ravel(), y.ravel())).T  # Original pixel coordinates
#     values = image.ravel()  # Corresponding pixel values
    
#     # Interpolate the pixel values at the actuator pixel coordinates
#     interpolated_intensities = griddata(points, values, pixel_coords, method='cubic')
    
#     return interpolated_intensities


# below is faster
def interpolate_pixel_intensities(image, pixel_coords):
    """
    Fast interpolation using scipy's map_coordinates, which is highly optimized for 2D grids.

    Args:
        image: 2D array of pixel intensities (image).
        pixel_coords: 2D array of actuator coordinates in pixel space.
        
    Returns:
        Interpolated intensities at the given actuator pixel coordinates.
    """
    # pixel_coords needs to be in the format [y_coords, x_coords] for map_coordinates
    pixel_coords = np.array(pixel_coords).T  # Transpose to match map_coordinates format

    # Perform the interpolation using map_coordinates (linear interpolation by default)
    interpolated_intensities = map_coordinates(image.T, pixel_coords, order=1, mode='nearest')
    
    return interpolated_intensities


def calibrate_transform_between_DM_and_image( dm_4_corners, img_4_corners , debug = False, fig_path= None ):

    stacked_corner_img = []
    img_corners = []
    corner_fits = {}
    for actuator_number, delta_img in zip(dm_4_corners, img_4_corners):  # <<< added actuator number 
        stacked_corner_img.append( delta_img )

        peak_pixels_raw = tuple( np.array( list(np.where( abs(delta_img) == np.max( abs(delta_img) )  ) ) ).ravel() )
                
        # fit actuator position in pixel space after interpolation and Gaussian fit 
        corner_fit_dict = interpolate_and_fit_gaussian(coord=peak_pixels_raw, radius=5, pixel_values= abs(delta_img), factor=5)
        corner_fits[actuator_number] = corner_fit_dict
        #plot_fit_results( corner_fit_dict )
        
        img_corners.append( ( corner_fit_dict['x0'],  corner_fit_dict['y0'] ) )
        # #Check individual registration          
        # plt.figure(actuator_number)
        # plt.imshow( delta_img )
        # plt.plot( corner_fit_dict['x0'],  corner_fit_dict['y0'], 'x', color='red', lw=4, label='registered position') 
        # plt.legend()
        # plt.colorbar(label='Intensity')
        # plt.show() 
    
        
    #[top_left, top_right, bottom_right, bottom_left]
    intersection = find_quadrilateral_diagonal_intersection( img_corners ) 
    
    # fig = plt.figure()
    # im = plt.imshow( np.sum( stacked_corner_img , axis=0 ))
    # cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
    # cbar.set_label(r'$\Delta$ Intensity', fontsize=15, labelpad=10)

    # for i,c in  enumerate( img_corners ):
    #     if i==0:
    #         plt.plot( c[0],  c[1], 'x', color='red', lw=4, label='registered position') 
    #     else:
    #         plt.plot( c[0],  c[1], 'x', color='red', lw=4 )
    
    # top_left, top_right, bottom_right, bottom_left = img_corners
    # plt.plot(  [top_left[0], bottom_right[0]], [top_left[1], bottom_right[1]] , 'r', lw=1)
    # plt.plot(  [bottom_left[0], top_right[0]],  [bottom_left[1], top_right[1]] , 'r', lw=1)
    # plt.plot( intersection[0], intersection[1], 'x' ,color='white', lw = 5 )
    # plt.legend()
    # savefig = fig_path + 'DM_center_in_pixel_space.png'
    # plt.savefig(savefig, dpi=300, bbox_inches = 'tight') 
    # plt.show()
    


    # Generate the DM coordinates for a 12x12 grid with missing corners with dictionaries
    # to convert between actuator number and coordinate frame 
    coords, dm_actuator_to_coord, dm_coord_to_actuator = generate_dm_coordinates()
        
    # get the coordinates of the DM corners used to probe the image space 
    dm_corners = np.array( [dm_actuator_to_coord[ c ] for c in dm_4_corners] )
    
    # ====== Fit the affine transformation to convert DM coordinates to Pixel space and visa versa
    # Important that indexing of corners in each space correspond!!!
    # ---!!!! --- THIS IS CRITICAL ---!!!----
    # ensure this by filling img_corners in the same order as dm_corners (as above).
    transform_matrix = fit_affine_transformation_with_center(dm_corners, img_corners, intersection)

    #FINALLY - projecting each actuator onto pixel space
    # get coordinates of each actuator 
    act_coord_list = np.array( [dm_actuator_to_coord[a] for a in range(140)] )
    ## check
    #plt.figure()
    #plt.scatter(act_coord_list[:, 0], act_coord_list[:, 1], color='blue', marker='o')
    #plt.show() 
    
    # put them in pixel space 
    pixel_coord_list = np.array( [dm_to_pixel(c, transform_matrix) for c in act_coord_list] )
    ## check 
    # plt.figure()
    # plt.scatter(pixel_coord_list[:, 0], pixel_coord_list[:, 1], color='blue', marker='o')
    # plt.show() 
    
    if debug:
        # Show the probing DM geometry 
        fig = plt.figure(1) 
        tmp =np.zeros(140)
        tmp[np.array(dm_4_corners)] = 1
        plt.imshow( get_DM_command_in_2D( tmp ) )
        if fig_path is not None:
            savefig = fig_path + 'DM_corner_poke_in_DM_space.png'
            fig.savefig(savefig, dpi=300, bbox_inches = 'tight' )


        # Show the actuator registration in Pixel space 
        fig = plt.figure(2)
        im = plt.imshow( np.sum( stacked_corner_img , axis=0 ))
        cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
        cbar.set_label(r'$\Delta$ Intensity', fontsize=15, labelpad=10)
        
        plt.scatter(pixel_coord_list[:, 0], pixel_coord_list[:, 1], color='blue', marker='.', label = 'DM actuators')
        
        for i,c in  enumerate( img_corners ):
            if i==0:
                plt.plot( c[0],  c[1], 'x', color='red', lw=4, label='registered corners') 
            else:
                plt.plot( c[0],  c[1], 'x', color='red', lw=4 )
        
        top_left, top_right,  bottom_left , bottom_right = sort_corners( img_corners ) 
        plt.plot(  [top_left[0],bottom_right[0]], [top_left[1],bottom_right[1]] , 'r', lw=1)
        plt.plot(  [bottom_left[0],top_right[0]],  [bottom_left[1],top_right[1]] , 'r', lw=1)
        plt.plot( intersection[0], intersection[1], 'x' ,color='white', lw = 5 )
        plt.legend()
        if fig_path is not None:
            savefig = fig_path + 'DM_registration_in_pixel_space.png'
            plt.savefig(savefig, dpi=300, bbox_inches = 'tight' )
        plt.show()

        

    ## write to file 
    write_dict = {
     "dm_corners" : dm_corners, 
     "img_corners" : img_corners, 
     "corner_fit_results" : corner_fits,
     "DM_center_pixel_space" : intersection,
     "actuator_to_pixel_matrix" : transform_matrix,
     "dm_actuator_to_coord" : dm_actuator_to_coord ,
     "actuator_coord_list_dm_space" : act_coord_list,
     "actuator_coord_list_pixel_space" : pixel_coord_list
    }
    
    return write_dict


    


if __name__== "__main__":
    
    # anyone else running this will need to change paths to their own. 
    
    example = 1 
    
    if example == 1:
        """
        doing DM registration from the interaction matrix 
        - example of how to deal with the case of different zonal basis (pinned edge actuators vs no pinning)
        """  
        
        # Define the parent directory where the search should start
        parent_dir = '/Users/bencb/Documents/baldr/data_sydney/A_FINAL_SYD_DATA_18-09-2024/tmp/'

        # Define the pattern to match
        pattern = os.path.join(parent_dir, '**/zonal_reconstructor/RECONSTRUCTORS_zonal*.fits')

        # Recursively search for files that match the pattern
        file_paths = glob.glob(pattern, recursive=True)

        dates = np.array( [datetime.datetime.strptime(f.split('.fits')[0].split('_')[-1],'%d-%m-%YT%H.%M.%S') for f in file_paths] )

        dm_center_in_detector = []
        for f in file_paths:
            
            with fits.open(f) as d:

                # ===================================
                # Look at DM center in pixel space 
                    
                # get inner corners for estiamting DM center in pixel space (have to deal seperately with pinned actuator basis)
                if d['I2M'].data.shape[1] == 100: # outer actuators are pinned, 
                    corner_indicies = get_inner_square_indices(outer_size=10, inner_offset=3, without_outer_corners=False)
                    
                elif d['I2M'].data.shape[1] == 140: # outer acrtuators are free 
                    print(140)
                    corner_indicies = get_inner_square_indices(outer_size=12, inner_offset=4, without_outer_corners=True)
                else:
                    print("CASE NOT MATCHED  d['I2M'].data.shape = { d['I2M'].data.shape}")
                    
                img_4_corners = []
                dm_4_corners = []
                for i in corner_indicies:
                    dm_4_corners.append( np.where( d['M2C'].data[:,i] )[0][0] )
                    #dm2px.get_DM_command_in_2D( d['M2C'].data[:,i]  # if you want to plot it 

                    tmp = np.zeros( d['I0'].data.shape )
                    tmp.reshape(-1)[d['pupil_pixels'].data] = d['IM'].data[i] 

                    #plt.imshow( tmp ); plt.show()
                    img_4_corners.append( abs(tmp ) )

                #plt.imshow( np.sum( tosee, axis=0 ) ); plt.show()

                # dm_4_corners should be an array of length 4 corresponding to the actuator index in the (flattened) DM command space
                # img_4_corners should be an array of length 4xNxM where NxM are the image dimensions.
                # !!! It is very important that img_4_corners are registered in the same order as dm_4_corners !!!
                transform_dict = calibrate_transform_between_DM_and_image( dm_4_corners, img_4_corners , debug=False, fig_path = None)

                dm_center_in_detector.append( transform_dict['DM_center_pixel_space'] )
                
    
    
    elif example == 2: 
        """
        #Example using poke ramp data to calibrate affine transform between DM and pixel space
        # change paths to whatever you want       
        """
        fig_path = '/Users/bencb/Documents/baldr/data_sydney/analysis_scripts/analysis_results/DM_registration_results/'
        files = glob.glob('/Users/bencb/Documents/baldr/data_sydney/A_FINAL_SYD_DATA_18-09-2024/tmp/09-09-2024/poke_ramp_data/pokeramp*.fits')

        if not os.path.exists( fig_path ):
            os.makedirs( fig_path )
            
        f = files[0]
        d = fits.open(f)
        No_ramps = int(d['SEQUENCE_IMGS'].header['#ramp steps'])
        max_ramp = float(d['SEQUENCE_IMGS'].header['in-poke max amp'])
        min_ramp = float(d['SEQUENCE_IMGS'].header['out-poke max amp'])
        ramp_values = np.linspace(min_ramp, max_ramp, No_ramps)

        #Nmodes_poked = int(d[0].header['HIERARCH Nmodes_poked'])
        Nact = int(d[0].header['HIERARCH Nact'])
        N0 = d['FPM_OUT'].data
        I0 = d['FPM_IN'].data

        poke_imgs = d['SEQUENCE_IMGS'].data[1:].reshape(No_ramps, Nact, I0.shape[0], I0.shape[1])

        dm_4_corners = get_inner_square_indices(outer_size=12, inner_offset=3) # flattened index of the DM actuator 
        ## look at what actuators we are considering
        # tmp =np.zeros(140)
        # tmp[np.array(dm_4_corners)] = 1
        # plt.figure()
        # plt.imshow( get_DM_command_in_2D( tmp ) )
        # plt.show() 
        
        img_4_corners = []
        for actuator_number in dm_4_corners:
            # for each actuator poke we get the corresponding (differenced) image
            # and append it to img_4_corners
            
            a = 4 # amplitude index 
            delta_img = poke_imgs[len(ramp_values)//2 + a][actuator_number] - I0 #poke_imgs[len(ramp_values)//2 - a][actuator_number]
            img_4_corners.append( delta_img  ) 
            
        # from the 4 images corresponding to the 4 (corner) actuator pokes we calibrate our transform
        
        # dm_4_corners should be an array of length 4 corresponding to the actuator index in the (flattened) DM command space
        # img_4_corners should be an array of length 4xNxM where NxM are the image dimensions.
        # !!! It is very important that img_4_corners are registered in the same order as dm_4_corners !!!
        transform_dict = calibrate_transform_between_DM_and_image( dm_4_corners, img_4_corners , debug=True, fig_path = fig_path)

        # example to overlay the registered actuators in pixel space with the pupil (FPM out of beam)
        fig = plt.figure(3)
        
        im = plt.imshow( N0 )
        cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
        cbar.set_label(r'Intensity', fontsize=15, labelpad=10)
        
        plt.scatter(transform_dict['actuator_coord_list_pixel_space'][:, 0],\
            transform_dict['actuator_coord_list_pixel_space'][:, 1], \
                color='blue', marker='.', label = 'DM actuators')
        
        plt.legend() 
        savefig = fig_path + 'pupil_on_DM_in_pixel_space.png'
        fig.savefig(savefig, dpi=300, bbox_inches = 'tight' )
        plt.show()    

        # write the transform dict to a json file 
        tstamp = datetime.datetime.now().strftime("%d-%m-%YT%H.%M.%S")
        serializable_dict = convert_to_serializable(transform_dict)

        with open(fig_path + f'DM2img_{tstamp}.json', 'w') as f:
            json.dump(serializable_dict, f)
            
            
        # example to interpolate the measured intensities onto registered actuators in pixel space 
        interpolated_intensities_I0 = interpolate_pixel_intensities(image = I0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])
        interpolated_intensities_N0 = interpolate_pixel_intensities(image = N0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

        plt.figure()
        plt.title('interpolating I0 onto \nregistered DM actuators')
        plt.imshow( get_DM_command_in_2D( interpolated_intensities_N0  ) )
        plt.show() 
        
        # example to register "well sensed" actuators 
        plt.figure()
        plt.title('interpolating I0 onto \nregistered DM actuators')
        plt.imshow( get_DM_command_in_2D( interpolated_intensities_N0  ) > 40 )
        plt.show() 




'''
Aperture/pupil utility functions largely copied from pyZELDA 
'''

import numpy as np
import scipy.ndimage as ndimage


def coordinates(dim, size, diameter=False, strict=False, center=(), cpix=False, normalized=True, outside=np.nan, polar=True):
    '''
    Returns rho,theta coordinates defining a circular aperture
    
    Parameters
    ----------
    dim : int
        Size of the output array
    
    size : int
        Size of the disk, representing either the radius (default) or the diameter
    
    diameter : bool, optional
        Specify if input size is the diameter. Default is 'False'
    
    strict : bool optional
        If set to Trye, size must be strictly less than (<), instead of less
        or equal (<=). Default is 'False'
    
    center : sequence, optional
        Specify the center of the disc. Default is '()', i.e. the center of the array
    
    cpix : bool optional
        If set to True, the disc is centered on pixel at position (dim//2, dim//2).
        Default is 'False', i.e. the disc is centered between 4 pixels
    
    normalized : bool optional
        Determines if the rho coordinates are normalized to 1. Default is True
    
    outside : float, optional
        Value used to fill the array outside of the aperture. Default is np.nan
    
    polar : bool, optional
        Return polar coordinates. Default is True
    
    Returns
    -------
    rho, theta : arrays
        Arrays containing the rho and theta coordinates of the aperture
    '''
    
    if (diameter is True):
        rad = size/2
    else:
        rad = size

    if (len(center) == 0):
        if (cpix is False):
            cx = (dim-1) / 2
            cy = (dim-1) / 2
        else:
            cx = dim // 2
            cy = dim // 2
    elif (len(center) == 2):
        cx = center[0]
        cy = center[1]
    else:
        raise ValueError('Error, you must pass 2 values for center')
        return None

    x = np.arange(dim, dtype=np.float64) - cx
    y = np.arange(dim, dtype=np.float64) - cy
    xx, yy = np.meshgrid(x, y)
    
    rho = np.sqrt(xx**2 + yy**2)
    theta = np.arctan2(yy, xx)
    
    if (strict is True):
        msk = (rho < rad)
    else:
        msk = (rho <= rad)

    xx[np.logical_not(msk)] = outside
    yy[np.logical_not(msk)] = outside
    rho[np.logical_not(msk)] = outside
    theta[np.logical_not(msk)] = outside

    if (normalized is True):
        xx  = xx / rad
        yy  = yy / rad
        rho = rho / rad

    if (polar is True):
        return rho, theta
    else:
        return xx, yy
    

def disc_obstructed(dim, size, obs, **kwargs):
    '''
    Create a numerical array containing a disc with central obstruction
    
    Parameters
    ----------
    dim : int
        Size of the output array
    
    size : int
        Size of the disk, representing either the radius (default) or the diameter
    
    obs : float
        Fractional size occupied by the central obstruction
    
    All other parameters are the same as for the disc() function and are described there.
    
    Returns
    -------
    disc : array
        An array containing a disc with central obstruction
    '''

    if (obs < 0) or (obs > 1):
        raise ValueError('obs value must be within [0,1]')
        return None
    
    ap_out = disc(dim, size, **kwargs)
    
    if obs == 0:
        return ap_out
    
    ap_in = disc(dim, size*obs, **kwargs)
    
    return ap_out - ap_in


def annulus(dim, inner_size, outer_size, **kwargs):
    '''
    Create a numerical array containing a disc with central obstruction
    
    Parameters
    ----------
    dim : int
        Size of the output array
    
    inner_size : int
        Inner size of the annulus, representing either the radius (default) or the diameter

    outer_size : int
        Outer size of the annulus, representing either the radius (default) or the diameter

    All other parameters are the same as for the disc() function and are described there.
    
    Returns
    -------
    disc : array
        An array containing a disc with central obstruction
    '''

    if (inner_size >= outer_size):
        raise ValueError('Inner radius must be smaller than outer radius')
        return None

    # outer ring
    ap_out = disc(dim, outer_size, **kwargs)

    # inner ring
    if inner_size <= 0:
        ap_final = ap_out
    else:
        ap_in = disc(dim, inner_size, **kwargs)
        ap_final = ap_out - ap_in
        
    return ap_final


def disc(dim, size, diameter=False, strict=False, center=(), cpix=False, invert=False, mask=False):
    '''
    Create a numerical array containing a disc.
    
    Parameters
    ----------
    dim : int
        Size of the output array
    
    size : int
        Size of the disk, representing either the radius (default) or the diameter
    
    diameter : bool, optional
        Specify if input size is the diameter. Default is 'False'
    
    strict : bool optional
        If set to Trye, size must be strictly less than (<), instead of less
        or equal (<=). Default is 'False'
    
    center : sequence, optional
        Specify the center of the disc. Default is '()', i.e. the center of the array
    
    cpix : bool optional
        If set to True, the disc is centered on pixel at position (dim//2, dim//2).
        Default is 'False', i.e. the disc is centered between 4 pixels
    
    invert : bool, optinal
        Specify if the disc must be inverted. Default is 'False'
    
    mask : bool, optional
        Specify if return value is a masked array or a numerical array. Default
        is 'False'
        
    Returns
    -------
    disc : array
        An array containing a disc with the specified parameters
    '''
    
    if (diameter is True):
        rad = size/2
    else:
        rad = size

    if (len(center) == 0):
        if (cpix is False):
            cx = (dim-1) / 2
            cy = (dim-1) / 2
        else:
            cx = dim // 2
            cy = dim // 2
    elif (len(center) == 2):
        cx = center[0]
        cy = center[1]
    else:
        raise ValueError('Error, you must pass 2 values for center')
        return None

    x = np.arange(dim, dtype=np.float64) - cx
    y = np.arange(dim, dtype=np.float64) - cy
    xx, yy = np.meshgrid(x, y)
    
    rr = np.sqrt(xx**2 + yy**2)

    if (strict is True):
        msk = (rr < rad)
    else:
        msk = (rr <= rad)
    
    if (invert is True):
        msk = np.logical_not(msk)

    if (mask is True):
        return msk
    else:
        d = np.zeros((dim, dim))
        d[msk] = 1
    
    return d


def _rotate_interp(array, alpha, center, mode='constant', cval=0):
    '''
    Rotation around a provided center

    This is the only way to be sure where exactly is the center of rotation.

    '''
    dtype = array.dtype
    dims  = array.shape
    alpha_rad = -np.deg2rad(alpha)

    x, y = np.meshgrid(np.arange(dims[1], dtype=dtype), np.arange(dims[0], dtype=dtype))

    xp = (x-center[0])*np.cos(alpha_rad) + (y-center[1])*np.sin(alpha_rad) + center[0]
    yp = -(x-center[0])*np.sin(alpha_rad) + (y-center[1])*np.cos(alpha_rad) + center[1]

    rotated = ndimage.map_coordinates(array, [yp, xp], mode=mode, cval=cval, order=3)
    
    return rotated


def _rotate_spider_interp(array, alpha0, center0, alpha1, center1):
    '''
    Rotation around a provided center

    This is the only way to be sure where exactly is the center of rotation.

    '''
    dtype = array.dtype
    dims  = array.shape
    alpha0_rad = -np.deg2rad(alpha0)
    alpha1_rad = -np.deg2rad(alpha1)

    x, y = np.meshgrid(np.arange(dims[1], dtype=dtype), np.arange(dims[0], dtype=dtype))

    x0 =  (x-center0[0])*np.cos(alpha0_rad) + (y-center0[1])*np.sin(alpha0_rad) + center0[0]
    y0 = -(x-center0[0])*np.sin(alpha0_rad) + (y-center0[1])*np.cos(alpha0_rad) + center0[1]

    x1 =  (x0-center1[0])*np.cos(alpha1_rad) + (y0-center1[1])*np.sin(alpha1_rad) + center1[0]
    y1 = -(x0-center1[0])*np.sin(alpha1_rad) + (y0-center1[1])*np.cos(alpha1_rad) + center1[1]
    
    rotated = ndimage.map_coordinates(array, [y1, x1], mode='constant', cval=0, order=3)
    
    return rotated


def vlt_pupil(dim, diameter, spiders_thickness=0.008, spiders_orientation=0, 
              dead_actuators=None,dead_actuator_diameter=0.025, strict=False, cpix=False):
    '''Very Large Telescope theoretical pupil with central obscuration and spiders

    Parameters
    ----------
    dim : int
        Size of the output array
    
    diameter : int
        Diameter the disk

    spiders_thickness : float
        Thickness of the spiders, in fraction of the pupil
        diameter. Default is 0.008

    spiders_orientation : float
        Orientation of the spiders. The zero-orientation corresponds
        to the orientation of the spiders when observing in ELEV
        mode. Default is 0

    dead_actuators : array
        Position of dead actuators in the pupil, given in fraction of
        the pupil size. The default values are for SPHERE dead
        actuators but any other values can be provided as a Nx2 array.

    dead_actuator_diameter : float
        Size of the dead actuators mask, in fraction of the pupil
        diameter. This is the dead actuators of SPHERE. Default is
        0.025

    strict : bool optional
        If set to Trye, size must be strictly less than (<), instead of less
        or equal (<=). Default is 'False'
    
    cpix : bool optional
        If set to True, the disc is centered on pixel at position (dim//2, dim//2).
        Default is 'False', i.e. the disc is centered between 4 pixels
    

    Returns
    -------
    pup : array
        An array containing a disc with the specified parameters

    '''

    # central obscuration (in fraction of the pupil)
    obs  = 1100/8000

    # spiders
    if spiders_thickness > 0:
        # adds some padding on the borders
        tdim = dim+50

        # dimensions
        cc = tdim // 2
        spdr = int(max(1, spiders_thickness*dim))
            
        ref = np.zeros((tdim, tdim))
        ref[cc:, cc:cc+spdr] = 1
        spider1 = _rotate_interp(ref, -5.5, (cc, cc+diameter/2))

        ref = np.zeros((tdim, tdim))
        ref[:cc, cc-spdr+1:cc+1] = 1
        spider2 = _rotate_interp(ref, -5.5, (cc, cc-diameter/2))
        
        ref = np.zeros((tdim, tdim))
        ref[cc:cc+spdr, cc:] = 1
        spider3 = _rotate_interp(ref, 5.5, (cc+diameter/2, cc))
        
        ref = np.zeros((tdim, tdim))
        ref[cc-spdr+1:cc+1, :cc] = 1
        spider4 = _rotate_interp(ref, 5.5, (cc-diameter/2, cc))

        spider0 = spider1 + spider2 + spider3 + spider4

        spider0 = _rotate_interp(spider1+spider2+spider3+spider4, 45+spiders_orientation, (cc, cc))
        
        spider0 = 1 - spider0
        spider0 = spider0[25:-25, 25:-25]
    else:
        spider0 = np.ones(dim)

    # main pupil
    pup = disc_obstructed(dim, diameter, obs, diameter=True, strict=strict, cpix=cpix)

    # add spiders
    pup *= spider0
    
    # dead actuators
    if (dead_actuator_diameter > 0) and (dead_actuators is not None):
        dead_actuators = np.array(dead_actuators)
        xarr = dead_actuators[0]
        yarr = dead_actuators[1]
        for i in range(len(xarr)):
            cx = xarr[i] * diameter + dim/2
            cy = yarr[i] * diameter + dim/2

            dead = disc(dim, dead_actuator_diameter*diameter, center=(cx, cy), invert=True)

            pup *= dead

    return (pup >= 0.5).astype(int)



def sphere_saxo_pupil(dim=240):
    '''SPHERE pupil in the SAXO geometry

    Parameters
    ----------
    dim : int
        Size of the output array. Default is 384
    
    Returns
    -------
    pup : array
        An array containing a disc with the specified parameters

    '''

    # fixed diameter
    diameter = 240

    if dim < diameter:
        raise ValueError('Image dimensions cannot be smaller than 384 pixels')
    
    # main pupil
    pup = disc_obstructed(diameter, diameter, 0.14, diameter=True, strict=False, cpix=False)
    
    return pup



    
    
    
def AT_pupil(dim, diameter, spiders_thickness=0.008, strict=False, cpix=False):
    '''Auxillary Telescope theoretical pupil with central obscuration and spiders
    
    function adapted from pyzelda..
    
    
    Parameters
    ----------
    dim : int
        Size of the output array (pixels)
    
    diameter : int
        Diameter the disk (pixels)
    spiders_thickness : float
        Thickness of the spiders, in fraction of the pupil
        diameter. Default is 0.008
    spiders_orientation : float
        Orientation of the spiders. The zero-orientation corresponds
        to the orientation of the spiders when observing in ELEV
        mode. Default is 0
    dead_actuators : array
        Position of dead actuators in the pupil, given in fraction of
        the pupil size. The default values are for SPHERE dead
        actuators but any other values can be provided as a Nx2 array.
    dead_actuator_diameter : float
        Size of the dead actuators mask, in fraction of the pupil
        diameter. This is the dead actuators of SPHERE. Default is
        0.025
    strict : bool optional
        If set to Trye, size must be strictly less than (<), instead of less
        or equal (<=). Default is 'False'
    
    cpix : bool optional
        If set to True, the disc is centered on pixel at position (dim//2, dim//2).
        Default is 'False', i.e. the disc is centered between 4 pixels
    
    Returns
    -------
    pup : array
        An array containing a disc with the specified parameters
    '''

    # central obscuration (in fraction of the pupil)
    obs  = 0.13/1.8
    spiders_orientation = 0

    pp1 = 2.5
    # spiders
    if spiders_thickness > 0:
        # adds some padding on the borders
        tdim = dim+50

        # dimensions
        cc = tdim // 2
        spdr = int(max(1, spiders_thickness*dim))
            
        ref = np.zeros((tdim, tdim))
        ref[cc:, cc:cc+spdr] = 1
        spider1 = _rotate_interp(ref, -pp1 , (cc, cc+diameter/2))

        ref = np.zeros((tdim, tdim))
        ref[:cc, cc-spdr+1:cc+1] = 1
        spider2 = _rotate_interp(ref, -pp1 , (cc, cc-diameter/2))
        
        ref = np.zeros((tdim, tdim))
        ref[cc:cc+spdr, cc:] = 1
        spider3 = _rotate_interp(ref, pp1 , (cc+diameter/2, cc))
        
        ref = np.zeros((tdim, tdim))
        ref[cc-spdr+1:cc+1, :cc] = 1
        spider4 = _rotate_interp(ref, pp1 , (cc-diameter/2, cc))

        spider0 = spider1 + spider2 + spider3 + spider4

        spider0 = _rotate_interp(spider1+spider2+spider3+spider4, 45+spiders_orientation, (cc, cc))
        
        spider0 = 1 - spider0
        spider0 = spider0[25:-25, 25:-25]
    else:
        spider0 = np.ones(dim)

    # main pupil
    pup = disc_obstructed(dim, diameter, obs, diameter=True, strict=strict, cpix=cpix)

    # add spiders
    pup *= spider0

    return (pup >= 0.5).astype(int)


def pick_pupil(pupil_geometry, dim, diameter ):
        
    if pupil_geometry == 'AT':
        pup = AT_pupil(dim = dim, diameter = diameter) 
    elif pupil_geometry == 'UT':
        pup = vlt_pupil(dim = dim, diameter =  diameter, dead_actuator_diameter=0) 
    elif pupil_geometry == 'disk':
        pup = disc( dim = dim, size = diameter//2) 
    else :
        print('no valid geometry defined (try pupil_geometry == disk, or UT, or AT\nassuming disk pupil')
        pup = disc( dim = dim, size = diameter//2) 

    return(pup)




import numpy as np #(version 2.1.1 works but incompatiple with numba)
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from scipy.stats import pearsonr
import pickle
from types import SimpleNamespace
from sklearn.linear_model import LinearRegression
import importlib # reimport package after edits: importlib.reload(bldr)
import os
import datetime
import scipy.interpolate as interpolate
# from courtney-barrer's fork of pyzelda
import pyzelda.zelda as zelda
import pyzelda.ztools as ztools
import pyzelda.utils.aperture as aperture
import pyzelda.utils.imutils as imutils
from baldrapp.common import phasescreens as ps
from baldrapp.common import utilities as util
from baldrapp.common import baldr_core as bldr
from baldrapp.common import DM_registration
from baldrapp.common import DM_basis

from baldrapp.common.baldr_core import StrehlModel 

import numpy as np
from sklearn.model_selection import train_test_split


import numpy as np
from sklearn.model_selection import train_test_split



# Function to load the model from a pickle file
def load_model_from_pickle(filename):
    """
    Loads the StrehlModel object from a pickle file.
    
    Args:
        filename (str): The file path from where the model should be loaded.
    
    Returns:
        StrehlModel: The loaded StrehlModel instance.
    """
    with open(filename, 'rb') as file:
        model = pickle.load(file)
        
    return model



class my_lin_fit:
    # Rows are samples, columns are features
    def __init__(self, model_name='pixelwise_first'):
        """
        Initialize the linear fit model.
        
        Parameters:
        - model_name: str, the name/type of model (currently supports 'pixelwise_first')
        """
        self.model_name = model_name
        self.models = None
        
    def fit(self, X, Y):
        """
        Fit the model based on the input features X and target Y.
        
        Parameters:
        - X: np.ndarray, shape (N, P), input data matrix (N samples, P features)
        - Y: np.ndarray, shape (N, P), target data matrix (same shape as X)
        
        Returns:
        - coe: list of model coefficients for each feature
        """
        if self.model_name == 'pixelwise_first':
            coe = []
            # Fit a first-order polynomial (linear) for each feature (each column)
            for v in range(X.shape[1]):
                coe.append(np.polyfit(X[:, v], Y[:, v], 1))  # Linear fit for each feature
            self.models = coe
            return coe 
        
    def apply(self, X):
        """
        Apply the fitted model to new input data X to make predictions.
        
        Parameters:
        - X: np.ndarray, input data for which to predict Y.
        
        Returns:
        - Y_pred: np.ndarray, predicted values based on the fitted models
        """
        if self.model_name == 'pixelwise_first':
            Y_pred = []
            # Apply the model to each feature
            for v in range(len(self.models)):
                a_i, b_i = self.models[v]
                if len(X.shape) == 1:
                    # X is 1D (single sample)
                    assert len(X) == len(self.models), "Dimension mismatch: X does not match model dimensions."
                    Y_pred.append(a_i * X[v] + b_i)
                elif len(X.shape) == 2:
                    # X is 2D (multiple samples)
                    assert X.shape[1] == len(self.models), "Dimension mismatch: X columns do not match model dimensions."
                    Y_pred.append(a_i * X[:, v] + b_i)
            return np.array(Y_pred).T  # Transpose to match the input shape
        else:
            return None
        


tstamp = datetime.datetime.now().strftime("%d-%m-%YT%H.%M.%S")

proj_path =  os.getcwd() #'/home/benja/Documents/BALDR/BaldrApp/' #'/home/rtc/Documents/BaldrApp/'

# initialize our ZWFS instrument
wvl0=1.25e-6
config_ini = proj_path  + '/baldrapp/configurations/BALDR_UT_J3.ini'#'/home/benja/Documents/BALDR/BaldrApp/configurations/BALDR_UT_J3.ini'
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=wvl0)


zwfs_ns.dm.actuator_coupling_factor = 0.9


fig_path = '/Users/bencb/Downloads/'#f'/home/benja/Downloads/act_cross_coupling_{zwfs_ns.dm.actuator_coupling_factor}_{tstamp}/' #f'/home/rtc/Documents/act_cross_coupling_{zwfs_ns.dm.actuator_coupling_factor}_{tstamp}/'
if os.path.exists(fig_path) == False:
    os.makedirs(fig_path) 
    
plot_intermediate_results = False


# set up detector class from zwfs_ns.detector 
# bldr_detector = bldr.detector( binning = (zwfs_ns.detector.binning, zwfs_ns.detector.binning), qe=zwfs_ns.detector.qe ,\
#     dit=zwfs_ns.detector.dit, ron= zwfs_ns.detector.ron)


# short hand for pupil dimensions (pixels)
#dim = zwfs_ns.grid.N * zwfs_ns.grid.padding_factor # should match zwfs_ns.pyZelda.pupil_dim
# spatial differential in pupil space 
dx = zwfs_ns.grid.D / zwfs_ns.grid.N
# get required simulation sampling rate to match physical parameters 
dt = dx * zwfs_ns.atmosphere.pixels_per_iteration / zwfs_ns.atmosphere.v # s # simulation sampling rate

print(f'current parameters have effective wind velocity = {round(zwfs_ns.atmosphere.v )}m/s')
scrn = ps.PhaseScreenKolmogorov(nx_size=zwfs_ns.grid.dim, 
                                pixel_scale=dx, 
                                r0=zwfs_ns.atmosphere.r0 * ( wvl0 /550e-9)**(6/5) , 
                                L0=zwfs_ns.atmosphere.l0, 
                                random_seed=1)

phase_scaling_factor = 0.3

# first stage AO 
basis_cropped = ztools.zernike.zernike_basis(nterms=150, npix=zwfs_ns.pyZelda.pupil_diameter)
# we have padding around telescope pupil (check zwfs_ns.pyZelda.pupil.shape and zwfs_ns.pyZelda.pupil_diameter) 
# so we need to put basis in the same frame  
basis_template = np.zeros( zwfs_ns.pyZelda.pupil.shape )
basis = np.array( [ util.insert_concentric( np.nan_to_num(b, 0), basis_template) for b in basis_cropped] )

#pupil_disk = basis[0] # we define a disk pupil without secondary - useful for removing Zernike modes later

Nmodes_removed = 14 # Default will be to remove Zernike modes 

# vibrations 
mode_indicies = [0, 1]
spectrum_type = ['1/f', '1/f']
opd = [50e-9, 50e-9]
vibration_frequencies = [15, 45] #Hz


# input flux scaling (photons / s / wavespace_pixel / nm) 
photon_flux_per_pixel_at_vlti = zwfs_ns.throughput.vlti_throughput * (np.pi * (zwfs_ns.grid.D/2)**2) / (np.pi * zwfs_ns.pyZelda.pupil_diameter/2)**2 * util.magnitude_to_photon_flux(magnitude=zwfs_ns.stellar.magnitude, band = zwfs_ns.stellar.waveband, wavelength= 1e9*wvl0)


# internal aberrations (opd in meters)
opd_internal = util.apply_parabolic_scratches(np.zeros( zwfs_ns.grid.pupil_mask.shape ) , dx=dx, dy=dx, list_a= [ 0.1], list_b = [0], list_c = [-2], width_list = [2*dx], depth_list = [100e-9])

opd_flat_dm = bldr.get_dm_displacement( command_vector= zwfs_ns.dm.dm_flat  , gain=zwfs_ns.dm.opd_per_cmd, \
                sigma= zwfs_ns.grid.dm_coord.act_sigma_wavesp, X=zwfs_ns.grid.wave_coord.X, Y=zwfs_ns.grid.wave_coord.Y,\
                    x0=zwfs_ns.grid.dm_coord.act_x0_list_wavesp, y0=zwfs_ns.grid.dm_coord.act_y0_list_wavesp )


# calculate reference (only with internal aberrations) optical gain (b0)
b0_wsp, _ = ztools.create_reference_wave_beyond_pupil_with_aberrations(opd_internal + opd_flat_dm , zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate, zwfs_ns.pyZelda.mask_Fratio,
                                       zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wvl0, clear=np.array([]), 
                                       sign_mask=np.array([]), cpix=False)

# # 
# b0_perfect, _ = ztools.create_reference_wave_beyond_pupil(zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate, zwfs_ns.pyZelda.mask_Fratio,
#                                        zwfs_ns.pyZelda.pupil_diameter, zwfs_ns.pyZelda.pupil, wvl0, clear=np.array([]), 
#                                        sign_mask=np.array([]), cpix=False)
# plt.figure(); plt.imshow( abs( b0_wsp ) - abs( b0_perfect) ) );plt.colorbar(); plt.show()
## >>>>>>> Note: the peak difference in b0 when including internal aberrations is 0.001. i.e. 0.008/0.8 = 1% difference in optical gain
    
    
# to put in pixel space (we just average with the same binning as the bldr detector)
b0 = bldr.average_subarrays( abs(b0_wsp) , (zwfs_ns.detector.binning, zwfs_ns.detector.binning) )
#plt.imshow( b0_pixelspace ); plt.colorbar(); plt.show()

# propagate through ZWFS to the detector plane intensities (in wavespace)
# phasemask in  
# I0_wsp =  ztools.propagate_opd_map( zwfs_ns.pyZelda.pupil * (opd_internal + opd_flat_dm ), zwfs_ns.pyZelda.mask_diameter, zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate,
#                                             zwfs_ns.pyZelda.mask_Fratio, zwfs_ns.pyZelda.pupil_diameter, photon_flux_per_pixel_at_vlti**0.5 *zwfs_ns.pyZelda.pupil, wave=zwfs_ns.optics.wvl0)

# # phasemask out
# N0_wsp =  ztools.propagate_opd_map(zwfs_ns.pyZelda.pupil * (opd_internal + opd_flat_dm ), zwfs_ns.pyZelda.mask_diameter, 0*zwfs_ns.pyZelda.mask_depth, zwfs_ns.pyZelda.mask_substrate,
#                                             zwfs_ns.pyZelda.mask_Fratio, zwfs_ns.pyZelda.pupil_diameter,  photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil, wave=zwfs_ns.optics.wvl0)

# # bin to detector pixelspace 
# I0 = bldr.detect( I0_wsp, binning = (zwfs_ns.detector.binning, zwfs_ns.detector.binning), qe=zwfs_ns.detector.qe , dit=zwfs_ns.detector.dit, ron= zwfs_ns.detector.ron, include_shotnoise=True, spectral_bandwidth = zwfs_ns.stellar.bandwidth )
# N0 = bldr.detect( N0_wsp, binning = (zwfs_ns.detector.binning, zwfs_ns.detector.binning), qe=zwfs_ns.detector.qe , dit=zwfs_ns.detector.dit, ron= zwfs_ns.detector.ron, include_shotnoise=True, spectral_bandwidth = zwfs_ns.stellar.bandwidth )


# quicker way - make sure get frames returns the same as the above!!!! 
I0 = bldr.get_I0( opd_input = 0 * zwfs_ns.pyZelda.pupil ,  amp_input = photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil, opd_internal=zwfs_ns.pyZelda.pupil * (opd_internal + opd_flat_dm), \
    zwfs_ns=zwfs_ns, detector=zwfs_ns.detector, include_shotnoise=True , use_pyZelda = True)

N0 = bldr.get_N0( opd_input = 0 * zwfs_ns.pyZelda.pupil ,  amp_input = photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil, opd_internal=zwfs_ns.pyZelda.pupil * (opd_internal + opd_flat_dm), \
    zwfs_ns=zwfs_ns, detector=zwfs_ns.detector, include_shotnoise=True , use_pyZelda = True)


# Build a basic Interaction matrix (IM) for the ZWFS
basis_name = 'Zonal_pinned_edges'
Nmodes = 100
M2C_0 = DM_basis.construct_command_basis( basis= basis_name, number_of_modes = Nmodes, without_piston=True).T  

### The amplitude input here is sqrt(photon flux)
zwfs_ns = bldr.classify_pupil_regions( opd_input = 0*opd_internal ,  amp_input = photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil , \
    opd_internal=opd_internal,  zwfs_ns = zwfs_ns , detector=zwfs_ns.detector , pupil_diameter_scaling = 1.0, pupil_offset = (0,0)) 



zwfs_ns = bldr.build_IM( zwfs_ns ,  calibration_opd_input = 0*opd_internal , calibration_amp_input = photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil  , \
            opd_internal = opd_internal,  basis = basis_name, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'double_sided_poke',\
                imgs_to_mean = 1, detector=zwfs_ns.detector)

# from IM register the DM in the detector pixelspace 
zwfs_ns = bldr.register_DM_in_pixelspace_from_IM( zwfs_ns, plot_intermediate_results=True  )


# build control matrices from IM 
#zwfs_ns = bldr.construct_ctrl_matricies_from_IM( zwfs_ns,  method = 'Eigen_TT-HO', Smax = 60, TT_vectors = DM_basis.get_tip_tilt_vectors() )

# or we create fit a linear zonal model to the IM
zwfs_ns = bldr.fit_linear_zonal_model( zwfs_ns, opd_internal, iterations = 100, photon_flux_per_pixel_at_vlti = photon_flux_per_pixel_at_vlti , \
    pearson_R_threshold = 0.6, phase_scaling_factor=phase_scaling_factor,   plot_intermediate=True , fig_path = None)


# example to interpolate any detected image onto the DM actuator grid
#DM_registration.interpolate_pixel_intensities(image = I0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space)

# add controllers 
#zwfs_ns = bldr.add_controllers_for_MVM_TT_HO( zwfs_ns, TT = 'PID', HO = 'leaky')
#zwfs_ns = bldr.add_controllers_for_MVM_TT_HO( zwfs_ns, TT = 'PID', HO = 'leaky')

if plot_intermediate_results:
    bldr.plot_eigenmodes( zwfs_ns , descr_label = f'dm_interactuator_coupling-{zwfs_ns.dm.actuator_coupling_factor}', save_path = fig_path )

### 
# DM registration 
###

# # get inner corners for estiamting DM center in pixel space (have to deal seperately with pinned actuator basis)
# if zwfs_ns.reco.IM.shape[0] == 100: # outer actuators are pinned, 
#     corner_indicies = DM_registration.get_inner_square_indices(outer_size=10, inner_offset=3, without_outer_corners=False)
    
# elif zwfs_ns.reco.IM.shape[0] == 140: # outer acrtuators are free 
#     print(140)
#     corner_indicies = DM_registration.get_inner_square_indices(outer_size=12, inner_offset=4, without_outer_corners=True)
# else:
#     print("CASE NOT MATCHED  d['I2M'].data.shape = { d['I2M'].data.shape}")
    
# img_4_corners = []
# dm_4_corners = []
# for i in corner_indicies:
#     dm_4_corners.append( np.where( M2C_0[i] )[0][0] )
#     #dm2px.get_DM_command_in_2D( d['M2C'].data[:,i]  # if you want to plot it 

#     tmp = np.zeros( zwfs_ns.pupil_regions.pupil_filt.shape )
#     tmp.reshape(-1)[zwfs_ns.pupil_regions.pupil_filt.reshape(-1)] = zwfs_ns.reco.IM[i] 

#     #plt.imshow( tmp ); plt.show()
#     img_4_corners.append( abs(tmp ) )

# #plt.imshow( np.sum( tosee, axis=0 ) ); plt.show()

# # dm_4_corners should be an array of length 4 corresponding to the actuator index in the (flattened) DM command space
# # img_4_corners should be an array of length 4xNxM where NxM are the image dimensions.
# # !!! It is very important that img_4_corners are registered in the same order as dm_4_corners !!!
# transform_dict = DM_registration.calibrate_transform_between_DM_and_image( dm_4_corners, img_4_corners, debug=plot_intermediate_results, fig_path = None )

# interpolated_intensities = DM_registration.interpolate_pixel_intensities(image = I0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])


# interpolate these fields onto the registered actuator grid
b0_dm = DM_registration.interpolate_pixel_intensities(image = b0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = I0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])
I0_dm = DM_registration.interpolate_pixel_intensities(image = I0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = b0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])
N0_dm = DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

#calibrate a model to map a subset of pixel intensities to Strehl Ratio 
#should eventually come back and debug for model_type = lin_comb - since it seemed to work better intially
strehl_model = bldr.calibrate_strehl_model( zwfs_ns, save_results_path = fig_path, train_fraction = 0.6, correlation_threshold = 0.6, \
   number_of_screen_initiations = 60, scrn_scaling_grid = np.logspace(-2, -0.5, 5), model_type = 'PixelWiseStrehlModel' ) #lin_comb') 

# or read one in  
# strehl_model_file = proj_path  + '/baldrapp/configurations/strehl_model_config-BALDR_UT_J3_2024-10-19T09.28.27.pkl'
# strehl_model = load_model_from_pickle(filename=strehl_model_file)




###
### CLOSED LOOP SIMULATION
#### 


### with zonal 

# optimize gains!!!
zonal_ctrl_dict = bldr.add_controllers_for_zonal_interp_no_projection( zwfs_ns ,  HO = 'PID' , return_controller = True) # HO = 'leaky'

amp_input =  photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil
dm_disturbance = np.zeros( 140 )

zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat.copy() 
zwfs_ns = bldr.reset_telemetry( zwfs_ns )

N0_dm = DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

kwargs = {"N0_dm":N0_dm, "HO_ctrl": zonal_ctrl_dict['HO_ctrl']  } 

Strehl_0_list = []
Strehl_1_list = []
Strehl_2_list = []
Strehl_est_list = []

static_aberrations_opd =  1e-7 * np.sum( [ b * a for a,b in zip( [0.5, 0.5, 0.2, 0.1, 0.1], basis[1:5])  ] , axis=0) 
print(  f'using { 1e9 * np.std(  static_aberrations_opd[zwfs_ns.pyZelda.pupil > 0.5] )}nm rms static aberrations' )
close_after = 5
iterations = 20

# open / close with strehl estimate 
# project out piston / tip/tilt
# optimize gains!!! 

improvement_ratio = {}
baldr_strehl_ratio = {}
for kp in [0, 0.1, 0.5, 1]:
    baldr_strehl_ratio[kp] = {}
    improvement_ratio[kp] = {}
    for ki in [0, 0.1 , 0.5, 0.9, 1]:
        print(f'   kp = {kp}, ki={ki}')

        #### SETUP #### 
        # same screen for all simulations
        scrn = ps.PhaseScreenKolmogorov(nx_size=zwfs_ns.grid.dim, 
                                        pixel_scale=dx, 
                                        r0=zwfs_ns.atmosphere.r0 * ( wvl0 /550e-9)**(6/5) , 
                                        L0=zwfs_ns.atmosphere.l0, 
                                        random_seed=1)

        #HO_ctrl.reset()
        zonal_ctrl_dict = bldr.add_controllers_for_zonal_interp_no_projection( zwfs_ns ,  HO = 'PID' , return_controller = True) # HO = 'leaky'
        # init all gains to 0

        amp_input =  photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil
        dm_disturbance = np.zeros( 140 )


        zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat.copy() 
        zwfs_ns = bldr.reset_telemetry( zwfs_ns )

        N0_dm = DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

        kwargs = {"N0_dm":N0_dm, "HO_ctrl": zonal_ctrl_dict['HO_ctrl']  } 

        Strehl_0_list = []
        Strehl_1_list = []
        Strehl_2_list = []
        Strehl_est_list = []

        #
        for it in range(iterations) :

            print( it )
            #print( kwargs['HO_ctrl'].integrals ) # <- this was the bug previously
            if it == close_after:

                kwargs["HO_ctrl"].kp = kp * np.ones( zonal_ctrl_dict['HO_ctrl'].kp.shape )
                kwargs["HO_ctrl"].ki = ki * np.ones( zonal_ctrl_dict['HO_ctrl'].ki.shape )
                #kwargs["HO_ctrl"].kd = 0 * np.ones( zonal_ctrl_dict['HO_ctrl'].kd.shape )
            
            # roll screen
            #for _ in range(10):
            scrn.add_row()
            
            # first stage AO
            if np.mod(it, 1) == 0: # only update the AO every few iterations to simulate latency 
                _ , reco_1 = bldr.first_stage_ao( scrn, Nmodes_removed , basis  , phase_scaling_factor = phase_scaling_factor, return_reconstructor = True )   
                
            ao_1 =  basis[0] * (phase_scaling_factor * scrn.scrn - reco_1)
            
            # opd after first stage AO
            opd_ao_1 = zwfs_ns.pyZelda.pupil * zwfs_ns.optics.wvl0 / (2*np.pi) * ao_1
            
            # add vibrations OPD and/or static aberrations
            opd_vibrations = static_aberrations_opd #np.zeros( ao_1.shape )
            
            # add BALDR DM OPD 
            opd_current_dm = bldr.get_dm_displacement( command_vector = zwfs_ns.dm.current_cmd   , gain=zwfs_ns.dm.opd_per_cmd, \
                        sigma = zwfs_ns.grid.dm_coord.act_sigma_wavesp, X=zwfs_ns.grid.wave_coord.X, Y=zwfs_ns.grid.wave_coord.Y,\
                            x0=zwfs_ns.grid.dm_coord.act_x0_list_wavesp, y0=zwfs_ns.grid.dm_coord.act_y0_list_wavesp )
            
            # sum all opd contributions in the Baldr input pupil plane 
            # set opd_ao_1 = 0 if rolling phasescreen on DM 
            bldr_opd_map = np.sum( [  opd_ao_1, opd_vibrations, opd_internal, opd_current_dm ] , axis=0 )
            bldr_opd_map -= np.mean( bldr_opd_map[zwfs_ns.pyZelda.pupil>0.5] ) # remove piston  
            
            ao_2 = zwfs_ns.pyZelda.pupil * (2*np.pi) / zwfs_ns.optics.wvl0  *  bldr_opd_map # phase radians 
            
            # get the real strehl ratios at various points (for tracking performance) 
            Strehl_0 = np.exp( - np.var( phase_scaling_factor * scrn.scrn[zwfs_ns.pyZelda.pupil>0.5]) ) # atmospheric strehl 
            Strehl_1 = np.exp( - np.var( ao_1[zwfs_ns.pyZelda.pupil>0.5]) ) # strehl after first stage AO 
            Strehl_2 = np.exp( - np.var( ao_2[zwfs_ns.pyZelda.pupil>0.5]) ) # strehl after baldr     

            
            i = bldr.AO_iteration( opd_input=bldr_opd_map, amp_input=amp_input, opd_internal=0*opd_internal, zwfs_ns=zwfs_ns, dm_disturbance = np.zeros(140),\
                record_telemetry=True, method='zonal_interp_no_projection', detector=zwfs_ns.detector, obs_intermediate_field=True, \
                    use_pyZelda = True, include_shotnoise=True, **kwargs)

            S_est = strehl_model.apply_model( np.array( [i / np.mean( N0[ strehl_model.detector_pupilmask ] )] ) ) 
            Strehl_est_list.append( S_est )
            
            if S_est < 0.1:
                zonal_ctrl_dict['HO_ctrl'].reset()
                zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat.copy()
           
            Strehl_0_list.append( Strehl_0 )
            Strehl_1_list.append( Strehl_1 )
            Strehl_2_list.append( Strehl_2 )

            print( round(Strehl_0,2), round(Strehl_1,2) , round(Strehl_2,2), 'S2 est ', S_est )

        baldr_strehl_ratio[kp][ki] =  Strehl_2_list
        improvement_ratio[kp][ki] =  np.array( [S2/ S1 for S1 , S2 in zip( Strehl_1_list[close_after:], Strehl_2_list[close_after:] ) ]  )



# Extract the keys for rows (kp) and columns (ki)
kp_values = sorted(improvement_ratio.keys())  # e.g., [1, 2, 3]
ki_values = sorted(improvement_ratio[kp_values[0]].keys())  # e.g., [1, 2, 3]

# Create a 2D numpy array
improvement_matrix_mean = np.array([[np.mean( improvement_ratio[kp][ki]) for ki in ki_values] for kp in kp_values])

plt.imshow( improvement_matrix_mean, cmap='viridis', origin='lower', extent=[ki_values[0], ki_values[-1], kp_values[0], kp_values[-1]], aspect='auto')
plt.colorbar()
plt.show() 

cax = plt.imshow(improvement_matrix_mean, cmap='YlGnBu', interpolation='nearest')

# Add color bar
plt.colorbar(cax)

# Labeling the axes
plt.xticks(np.arange(len(ki_values)), ki_values)  # x-axis ticks
plt.yticks(np.arange(len(kp_values)), kp_values)  # y-axis ticks
plt.xlabel('ki')
plt.ylabel('kp')

# Add a title
plt.title('Improvement Ratio Heatmap')
plt.show() 




# impact of actuator cross coupling on correction performance
# s




#### SETUP #### 
# same screen for all simulations
scrn = ps.PhaseScreenKolmogorov(nx_size=zwfs_ns.grid.dim, 
                                pixel_scale=dx, 
                                r0=zwfs_ns.atmosphere.r0 * ( wvl0 /550e-9)**(6/5) , 
                                L0=zwfs_ns.atmosphere.l0, 
                                random_seed=1)

#HO_ctrl.reset()
zonal_ctrl_dict = bldr.add_controllers_for_zonal_interp_no_projection( zwfs_ns ,  HO = 'PID' , return_controller = True) # HO = 'leaky'
# init all gains to 0

amp_input =  photon_flux_per_pixel_at_vlti**0.5 * zwfs_ns.pyZelda.pupil
dm_disturbance = np.zeros( 140 )


zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat.copy() 
zwfs_ns = bldr.reset_telemetry( zwfs_ns )

N0_dm = DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = zwfs_ns.dm2pix_registration.actuator_coord_list_pixel_space) #DM_registration.interpolate_pixel_intensities(image = N0, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])


static_aberrations_opd =  1e-7 * np.sum( [ b * a for a,b in zip( [0.5, 0.5, 0.2, 0.1, 0.1], basis[1:5])  ] , axis=0) 
print(  f'using { 1e9 * np.std(  static_aberrations_opd[zwfs_ns.pyZelda.pupil > 0.5] )}nm rms static aberrations' )
close_after = 5
iterations = 20

kwargs = {"N0_dm":N0_dm, "HO_ctrl": zonal_ctrl_dict['HO_ctrl']  } 

Strehl_0_list = []
Strehl_1_list = []
Strehl_11_list = []
Strehl_2_list = []
Strehl_est_list = []

kp = 0.0
ki = 0.4 

#
for it in range(iterations) :

    print( it )
    #print( kwargs['HO_ctrl'].integrals ) # <- this was the bug previously
    if it == close_after:

        kwargs["HO_ctrl"].kp = kp * np.ones( zonal_ctrl_dict['HO_ctrl'].kp.shape )
        kwargs["HO_ctrl"].ki = ki * np.ones( zonal_ctrl_dict['HO_ctrl'].ki.shape )
        #kwargs["HO_ctrl"].kd = 0 * np.ones( zonal_ctrl_dict['HO_ctrl'].kd.shape )
    
    # roll screen
    #for _ in range(10):
    scrn.add_row()
    
    # first stage AO
    if np.mod(it, 1) == 0: # only update the AO every few iterations to simulate latency 
        _ , reco_1 = bldr.first_stage_ao( scrn, Nmodes_removed , basis  , phase_scaling_factor = phase_scaling_factor, return_reconstructor = True )   
        
    ao_1 =  basis[0] * (phase_scaling_factor * scrn.scrn - reco_1)
    
    # opd after first stage AO
    opd_ao_1 = zwfs_ns.pyZelda.pupil * zwfs_ns.optics.wvl0 / (2*np.pi) * ao_1
    
    # add vibrations OPD and/or static aberrations
    opd_vibrations = static_aberrations_opd #np.zeros( ao_1.shape )
    

    rad_before_baldr_dm = (2*np.pi) / zwfs_ns.optics.wvl0  * np.sum( [  opd_ao_1, opd_vibrations, opd_internal ] , axis=0 )

    # add BALDR DM OPD 
    opd_current_dm = bldr.get_dm_displacement( command_vector = zwfs_ns.dm.current_cmd   , gain=zwfs_ns.dm.opd_per_cmd, \
                sigma = zwfs_ns.grid.dm_coord.act_sigma_wavesp, X=zwfs_ns.grid.wave_coord.X, Y=zwfs_ns.grid.wave_coord.Y,\
                    x0=zwfs_ns.grid.dm_coord.act_x0_list_wavesp, y0=zwfs_ns.grid.dm_coord.act_y0_list_wavesp )
    
    # sum all opd contributions in the Baldr input pupil plane 
    # set opd_ao_1 = 0 if rolling phasescreen on DM 
    bldr_opd_map = np.sum( [  opd_ao_1, opd_vibrations, opd_internal, opd_current_dm ] , axis=0 )
    bldr_opd_map -= np.mean( bldr_opd_map[zwfs_ns.pyZelda.pupil>0.5] ) # remove piston  
    
    ao_2 = zwfs_ns.pyZelda.pupil * (2*np.pi) / zwfs_ns.optics.wvl0  *  bldr_opd_map # phase radians 
    
    # get the real strehl ratios at various points (for tracking performance) 
    Strehl_0 = np.exp( - np.var( phase_scaling_factor * scrn.scrn[zwfs_ns.pyZelda.pupil>0.5]) ) # atmospheric strehl 
    Strehl_1 = np.exp( - np.var( ao_1[zwfs_ns.pyZelda.pupil>0.5]) ) # strehl after first stage AO 
    Strehl_11 = np.exp( - np.var( rad_before_baldr_dm[zwfs_ns.pyZelda.pupil>0.5]  ) )
    Strehl_2 = np.exp( - np.var( ao_2[zwfs_ns.pyZelda.pupil>0.5]) ) # strehl after baldr     

    
    i = bldr.AO_iteration( opd_input=bldr_opd_map, amp_input=amp_input, opd_internal=0*opd_internal, zwfs_ns=zwfs_ns, dm_disturbance = np.zeros(140),\
        record_telemetry=True, method='zonal_interp_no_projection', detector=zwfs_ns.detector, obs_intermediate_field=True, \
            use_pyZelda = True, include_shotnoise=True, **kwargs)

    S_est = strehl_model.apply_model( np.array( [i / np.mean( N0[ strehl_model.detector_pupilmask ] )] ) ) 
    Strehl_est_list.append( S_est )
    
    if S_est < 0.1:
        zonal_ctrl_dict['HO_ctrl'].reset()
        zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat.copy()

    Strehl_0_list.append( Strehl_0 )
    Strehl_1_list.append( Strehl_1 )
    Strehl_11_list.append( Strehl_11 )
    Strehl_2_list.append( Strehl_2 )

    print( round(Strehl_0,2), round(Strehl_1,2) , round(Strehl_11,2) , round(Strehl_2,2), 'S2 est ', S_est )





i = -1
#im_dm_dist = np.array( [util.get_DM_command_in_2D( a ) for a in zwfs_ns.telem.dm_disturb_list] )
im_phase = np.array( [util.get_DM_command_in_2D(a) for a  in zwfs_ns.telem.i_dm_list ] ) # zwfs_ns.telem.field_phase ) 
im_int = np.array( zwfs_ns.telem.i_list  ) 
im_cmd = np.array( [util.get_DM_command_in_2D( a ) for a in (np.array(zwfs_ns.telem.c_TT_list) + np.array(zwfs_ns.telem.c_HO_list)  ) ] )


#line_x = np.linspace(0, i, i)
# line_eHO = np.array( zwfs_ns.telem.e_HO_list ) [:i]
# line_eTT = np.array( zwfs_ns.telem.e_TT_list )[:i]
# line_S = np.array( zwfs_ns.telem.strehl )[:i]
# line_rmse = np.array( zwfs_ns.telem.rmse_list )[:i]
line_eHO = np.array( zwfs_ns.telem.e_HO_list ) 
line_eTT = np.array( zwfs_ns.telem.e_TT_list )
line_S = np.array( zwfs_ns.telem.strehl )
line_rmse = np.array( zwfs_ns.telem.rmse_list )

# Define plot data
#image_list =  [im_phase[-1], im_phase[-1], im_int[-1], im_cmd[-1]]
image_list =  [ zwfs_ns.telem.field_phase, im_phase, im_int, im_cmd]
image_title_list =  ['DM disturbance', 'input phase', 'intensity', 'reco. command']
image_colorbar_list = ['DM units', 'radians', 'adu', 'DM units']

plot_list = [ line_eHO, line_eTT, line_S, line_rmse ] 
plot_ylabel_list = ['e_HO', 'e_TT', 'Strehl', 'rmse']
plot_xlabel_list = ['iteration' for _ in plot_list]
plot_title_list = ['' for _ in plot_list]

#vlims = [(0, 1), (0, 1), (0, 1)]  # Set vmin and vmax for each image

util.create_telem_mosaic([a[-1] for a in image_list], image_title_list, image_colorbar_list, 
                plot_list, plot_title_list, plot_xlabel_list, plot_ylabel_list)

util.display_images_with_slider(image_lists = image_list)
       

# check with actuators go bad 
tmpcmd = np.zeros(140)
tmpcmd[zwfs_ns.reco.linear_zonal_model.act_filt_recommended] = zwfs_ns.telem.e_HO_list[-1]
plt.imshow( util.get_DM_command_in_2D( tmpcmd ) )
plt.imshow( util.get_DM_command_in_2D( tmpcmd ) ); plt.show()







#!/usr/bin/env python
from setuptools import setup, find_packages
import os

# Read the requirements from requirements.txt
def read_requirements():
    requirements_path = os.path.join(os.path.dirname(__file__), "requirements.txt")
    with open(requirements_path) as req_file:
        requirements = req_file.read().splitlines()
        # Exclude git+ lines from install_requires
        return [req for req in requirements if not req.startswith("git+")]

# Extract dependency links for git-based dependencies
def read_dependency_links():
    requirements_path = os.path.join(os.path.dirname(__file__), "requirements.txt")
    with open(requirements_path) as req_file:
        requirements = req_file.read().splitlines()
        # Only include git+ lines
        return [req for req in requirements if req.startswith("git+")]

setup(
    name="BaldrApp",
    version="0.1.7",
    description="Simulating Baldr - the Zernike Wavefront Sensor for Asgard",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Benjamin Courtney-Barrer",
    author_email="benjamin.courtney-barrer@anu.edu.au",
    url="https://github.com/your_username/your_project",
    package_data={
        "baldrapp.configurations": ["*"],  # Include all files in this directory
        "baldrapp.data": ["*"],  # Include all files in this directory
    },
    packages=find_packages(include=["baldrapp", "baldrapp.*"]),  # Include baldrapp and subpackages
    scripts=[
        "baldrapp/apps/baldr_closed_loop_app/closed_loop_pyqtgraph.py",
        "baldrapp/apps/baldr_alignment_app/Baldr_Fresnel_App.py"
    ],
    install_requires=read_requirements(),  # Include requirements dynamically
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    license="MIT",
    license_files=["LICENSE"],
)